import json
import logging
from functools import cached_property
from typing import Any, Generator, Optional

import boto3
from botocore.exceptions import ClientError, NoCredentialsError

from strong_opx.project import Project
from strong_opx.providers.aws.config import get_aws_config
from strong_opx.providers.aws.errors import handle_boto_error
from strong_opx.providers.aws.iam import get_current_account_id
from strong_opx.providers.docker_registry import AbstractDockerRegistry
from strong_opx.utils.shell import shell

logger = logging.getLogger(__name__)


class DockerRegistry(AbstractDockerRegistry):
    @cached_property
    def client(self):
        return boto3.client("ecr")

    def login(self):
        account_id = get_current_account_id()
        region = get_aws_config("region")

        docker_executable = Project.current().config.docker_executable

        shell(
            f"aws ecr get-login-password --region {region} | "
            f"{docker_executable} login --username AWS --password-stdin "
            f"{account_id}.dkr.ecr.{region}.amazonaws.com",
            shell=True,
        )

    def create_repository(self, repository_name: str) -> dict[str, Any]:
        logger.info(f"Creating ECR repository: {repository_name}...")
        response = self.client.create_repository(repositoryName=repository_name)
        self.client.put_lifecycle_policy(
            repositoryName=repository_name,
            lifecyclePolicyText=json.dumps(
                {
                    "rules": [
                        {
                            "rulePriority": 1,
                            "description": f"Only keep {self.max_images_to_keep} images for {self.environment.name}",
                            "selection": {
                                "tagStatus": "tagged",
                                "tagPrefixList": [f"{self.environment.name}-"],
                                "countType": "imageCountMoreThan",
                                "countNumber": self.max_images_to_keep,
                            },
                            "action": {"type": "expire"},
                        }
                    ]
                }
            ),
        )

        return response["repository"]["repositoryUri"]

    def get_repository_uri(self, repository_name: str) -> Optional[str]:
        try:
            response = self.client.describe_repositories(repositoryNames=[repository_name])
        except (ClientError, NoCredentialsError) as e:
            handle_boto_error(e)
        else:
            return response["repositories"][0]["repositoryUri"]

    def iter_image_tags(self, repository_name: str) -> Generator[str, None, None]:
        try:
            paginator = self.client.get_paginator("describe_images")
            for page in paginator.paginate(repositoryName=repository_name):
                for image in page["imageDetails"]:
                    for tag in image.get("imageTags", []):
                        yield tag
        except (ClientError, NoCredentialsError) as e:
            handle_boto_error(e, ignore=("RepositoryNotFoundException",))



import re
from typing import TextIO

from strong_opx.utils.tracking import OpxString, Position, set_position

VARIABLE_NAME_RE = re.compile(r'variable\s+\"([^"]+)\"\s+{')
VARIABLE_DEFAULT_RE = re.compile(r"default\s+=")


class FileReader:
    def __init__(self, file_path: str, stream: TextIO):
        self.file_path = file_path
        self.stream = stream

        self.line = 1
        self.column = 1

    def peak(self, n: int) -> str:
        s = self.stream.read(n)
        self.stream.seek(self.stream.tell() - len(s), 0)
        return s

    def read(self, n: int) -> str:
        s = self.stream.read(n)
        breaks = s.count("\n")
        if breaks:
            self.line += breaks
            self.column = len(s) - s.rindex("\n")
        else:
            self.column += len(s)

        return s

    def discard_whitespaces(self):
        while self.peak(1).isspace():
            self.read(1)

    def read_comment(self) -> str:
        content = ""
        while True:
            s = self.read(1)
            if not s or s == "\n":
                break

            content += s

        return content

    def read_string(self, end: str) -> str:
        content = ""
        while True:
            s = self.read(1)
            if not s:
                break

            if s == end and content[-1:] != "\\":
                break

            content += s

        return content

    def read_until(self, end: str) -> str:
        content = ""
        while True:
            s = self.read(1)
            if not s:
                break

            if s == "#":
                self.read_comment()
                continue

            content += s
            if s in "'\"":
                content += self.read_string(s) + s

            if s == end:
                break

        return content

    def read_block(self, start: str, end: str) -> str:
        stack = 1
        content = ""
        while True:
            s = self.read(1)
            if not s:
                break

            content += s
            if s == "#":
                content += self.read_comment() + "\n"
                continue

            if s in "'\"":
                content += self.read_string(s) + s
                continue

            if s == start:
                stack += 1
            elif s == end:
                stack -= 1
                if not stack:
                    break

        return content


class HCLVariableExtractor:
    def __init__(self):
        self.required_vars: set[OpxString] = set()
        self.optional_vars: set[OpxString] = set()

    def extract(self, file_path: str, stream: TextIO) -> None:
        reader = FileReader(file_path, stream)

        while True:
            reader.discard_whitespaces()
            initial_line = reader.line
            initial_col = reader.column

            block_prefix = reader.read_until("{")
            if not block_prefix:
                break

            block_content = reader.read_block("{", "}")
            match = VARIABLE_NAME_RE.match(block_prefix)
            if not match:
                continue

            var_name = OpxString(match.group(1))
            start_offset, end_offset = match.span(1)
            set_position(
                var_name,
                file_path,
                Position.from_offset(block_prefix, start_offset, initial_line, initial_col),
                Position.from_offset(block_prefix, end_offset, initial_line, initial_col),
            )

            if VARIABLE_DEFAULT_RE.search(block_content):
                self.optional_vars.add(var_name)
            else:
                self.required_vars.add(var_name)



import argparse
from typing import Any, Optional

import boto3
from botocore.exceptions import ClientError
from colorama import Fore as TextColor

from strong_opx.config import system_config
from strong_opx.exceptions import CommandError
from strong_opx.management.command import BaseCommand
from strong_opx.providers.aws import AWSCredentialConfig
from strong_opx.providers.aws.credentials import AWSCredential, mfa_profile_name


def validate_mfa_token(token: str) -> str:
    if len(token) != 6:
        raise argparse.ArgumentTypeError("MFA Token must be 6 digits")

    return token


class Command(BaseCommand):
    help_text = "Refresh AWS Security Token for MFA enabled profiles"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument("mfa_token", type=validate_mfa_token, help="OTP from MFA device")
        parser.add_argument("--profile", help="AWS Profile Name", required=True)
        parser.add_argument(
            "--duration",
            type=int,
            default=43200,
            help="The duration, in seconds, that the credentials should remain valid. Default 12 hrs",
        )

    def handle(
        self,
        mfa_token: str,
        profile: Optional[str],
        duration: int,
        **options: Any,
    ) -> None:
        if profile is None:
            profile = system_config.get("aws", "aws_profile", fallback="default")

        aws_config = AWSCredentialConfig()
        section_name = mfa_profile_name(profile)

        if not aws_config.has_section(section_name):
            raise CommandError(f"No MFA configuration found for profile {profile}")

        aws_mfa_device = aws_config.get(section_name, "aws_mfa_device")
        sts = boto3.Session(profile_name=section_name).client("sts")

        try:
            response = sts.get_session_token(
                DurationSeconds=duration,
                SerialNumber=aws_mfa_device,
                TokenCode=mfa_token,
            )["Credentials"]

            credentials = AWSCredential(
                aws_access_key_id=response["AccessKeyId"],
                aws_secret_access_key=response["SecretAccessKey"],
                aws_session_token=response["SessionToken"],
            )
            expiration = response["Expiration"]
        except ClientError as e:
            if e.operation_name == "GetSessionToken":
                raise CommandError(e.response.get("Error", {}).get("Message", "Unknown"))

            raise

        aws_config.set_credential(profile, credentials, expiration)
        print(f"{TextColor.GREEN}AWS Security Token refreshed successfully{TextColor.RESET}")



import configparser
import os


class Config(configparser.ConfigParser):
    def __init__(self, config_path: str):
        self.config_path = config_path
        super().__init__()

        self.read([self.config_path])

    def save(self) -> None:
        os.makedirs(os.path.dirname(self.config_path), exist_ok=True)
        with open(self.config_path, "w") as f:
            self.write(f)


class SystemConfig(Config):
    def set(self, section: str, *args, **kwargs) -> None:
        if section not in self:
            self.add_section(section)

        super().set(section, *args, **kwargs)

    def register_project(self, name: str, path: str):
        if "projects" not in self:
            self.add_section("projects")

        self.set("projects", name, path)
        self.save()

    def unregister_project(self, name: str):
        if "projects" not in self:
            self.add_section("projects")

        self.remove_option("projects", name)
        self.save()

    @property
    def registered_projects(self) -> list[str]:
        projects = []

        if "projects" not in self:
            return projects

        return list(self["projects"].keys())

    def get_project_path(self, name: str) -> str:
        return self.get("projects", name)

    def get_project_config_dir(self, project_name: str) -> str:
        return os.path.join(os.path.dirname(self.config_path), project_name)

    def get_project_config(self, project_name: str) -> Config:
        project_config_path = os.path.join(self.get_project_config_dir(project_name), "config")
        return Config(project_config_path)



import unittest
from unittest import mock

from strong_opx.exceptions import VaultError
from strong_opx.vault import VaultCipher


class VaultCipherTests(unittest.TestCase):
    secret = "some-secret"
    plain_text = "some-value"

    def test_encrypt_decrypt(self):
        cipher = VaultCipher.encrypt(self.plain_text, self.secret)
        decrypted_valued = cipher.decrypt(self.secret)
        self.assertEqual(self.plain_text, decrypted_valued)

    def test_parse(self):
        cipher_envelope = str(VaultCipher.encrypt(self.plain_text, self.secret))
        parsed_cipher = VaultCipher.parse(cipher_envelope)

        self.assertEqual(str(parsed_cipher), cipher_envelope)

    def test_decrypt_wrong_secret(self):
        cipher = VaultCipher.encrypt(self.plain_text, self.secret)
        with self.assertRaises(VaultError) as cm:
            cipher.decrypt("wrong-secret")

        self.assertEqual(str(cm.exception), "Decryption failed. Did you copied from other environment?")

    def test_str(self):
        cipher = VaultCipher.encrypt(self.plain_text, self.secret)

        cipher_envelope = str(cipher)
        self.assertTrue(cipher_envelope.startswith("$STRONG_OPX_VAULT;1.0;AES256\n"))

    def test_str__lines_should_be_less_than_80_chars(self):
        cipher = VaultCipher.encrypt(self.plain_text, self.secret)

        cipher_envelope = str(cipher)
        for line in cipher_envelope.splitlines():
            self.assertLessEqual(len(line), 80)

    @mock.patch("strong_opx.project.Project")
    def test_call_should_resolve_secret_from_project(self, project_mock: mock.Mock):
        project_mock.current.return_value.selected_environment.vault_secret = self.secret

        cipher = VaultCipher.encrypt(self.plain_text, self.secret)
        self.assertEqual(cipher(), self.plain_text)



import os
from typing import Optional

from pydantic.dataclasses import dataclass

from strong_opx.exceptions import ImproperlyConfiguredError


@dataclass
class AWSConfig:
    region: Optional[str] = None

    def dict(self) -> dict[str, str]:
        d = {}
        if self.region:
            d["AWS_REGION"] = self.region

        return d


def get_aws_config(var: str, required=True) -> Optional[str]:
    var_name = f"AWS_{var.upper()}"
    value = os.getenv(var_name)
    if value is None and required:
        raise ImproperlyConfiguredError(
            f"AWS {var.capitalize()} is not configured. "
            f"Either configure that in project or environment under aws.{var.lower()}"
        )

    return value



import logging
from functools import cached_property
from typing import TYPE_CHECKING, Generator, Optional

from google.api_core.exceptions import NotFound
from google.cloud.artifactregistry_v1 import (
    ArtifactRegistryClient,
    CleanupPolicy,
    CleanupPolicyMostRecentVersions,
    CreateRepositoryRequest,
    ListVersionsRequest,
    Repository,
    VersionView,
)

from strong_opx.exceptions import ProcessError, RepositoryNotFoundException
from strong_opx.providers.discovery import current_provider
from strong_opx.providers.docker_registry import AbstractDockerRegistry
from strong_opx.utils.shell import shell

if TYPE_CHECKING:
    from strong_opx.providers.gcloud.provider import GCloudProvider

logger = logging.getLogger(__name__)


def parse_repository_name(name: str) -> tuple[str, str]:
    parts = name.split("/")
    if len(parts) != 2:
        raise ValueError(f"Invalid repository name: {name}. It should be of format: <repository-name>/<package-name>")

    return parts[0], parts[1]


class DockerRegistry(AbstractDockerRegistry):
    @cached_property
    def provider(self) -> "GCloudProvider":
        provider: "GCloudProvider" = current_provider()  # type: ignore[assignment]
        return provider

    @cached_property
    def client(self):
        return ArtifactRegistryClient()

    def login(self):
        shell(
            [
                "gcloud",
                "auth",
                "configure-docker",
                "--quiet",
                f"{self.provider.config.compute_region}-docker.pkg.dev",
            ]
        )

    def create_repository(self, full_repository_name: str) -> str:
        repository_name, package_name = parse_repository_name(full_repository_name)
        logger.info(f"Creating Artifact repository: {repository_name}...")

        request = CreateRepositoryRequest(
            parent=self.provider.gcp_project_path,
            repository_id=repository_name,
            repository=Repository(
                format_=Repository.Format.DOCKER,
                labels={"provisioned-by": "strong-opx"},
                cleanup_policies={
                    "keep-recent-versions": CleanupPolicy(
                        action=CleanupPolicy.Action.KEEP,
                        most_recent_versions=CleanupPolicyMostRecentVersions(
                            keep_count=self.max_images_to_keep,
                        ),
                    )
                },
            ),
        )

        self.client.create_repository(request=request)
        return self._repository_uri(repository_name, package_name)

    def get_repository_uri(self, full_repository_name: str) -> Optional[str]:
        repository_name, package_name = parse_repository_name(full_repository_name)

        repo_path = "/".join([self.provider.gcp_project_path, f"repositories/{repository_name}"])

        try:
            repository = self.client.get_repository(name=repo_path)
            if repository.format_ != Repository.Format.DOCKER:
                raise ProcessError(f"Repository {repository_name} is not a Docker repository.")
        except NotFound:
            raise RepositoryNotFoundException()

        return self._repository_uri(repository_name, package_name)

    def _repository_uri(self, repository_name: str, package_name: str) -> str:
        config = self.provider.config
        return "/".join(
            [
                f"{config.compute_region}-docker.pkg.dev",
                config.project,
                repository_name,
                package_name,
            ]
        )

    def iter_image_tags(self, full_repository_name: str) -> Generator[str, None, None]:
        repository_name, package_name = parse_repository_name(full_repository_name)
        package_path = "/".join(
            [
                self.provider.gcp_project_path,
                f"repositories/{repository_name}",
                f"packages/{package_name}",
            ]
        )

        request = ListVersionsRequest(parent=package_path, page_size=100, view=VersionView.FULL)

        try:
            versions = self.client.list_versions(request=request)
            for version in versions:
                for tag in version.related_tags:
                    yield tag.name.rsplit("/", 1)[1]

        except NotFound:
            pass



import os

from strong_opx.codegen.generator import TemplateGenerator
from strong_opx.codegen.questions import ChoiceQuestion, SimpleQuestion
from strong_opx.project import Environment
from strong_opx.template import Context


class Generator(TemplateGenerator):
    template_dir = "environment"

    questions = {
        "environment_name": SimpleQuestion(
            prompt="Specify environment name",
            validation_re=r"[A-Za-z][A-Za-z0-9-]",
        ),
        "aws_region": ChoiceQuestion(
            prompt="Specify AWS region",
            choices=[
                "us-east-1",
                "us-east-2",
                "us-west-1",
                "us-west-2",
            ],
        ),
    }

    def get_output_dir(self, context: "Context") -> str:
        return os.path.join(self.project.path, "environments", context["environment_name"])

    def get_context(self, context: "Context") -> "Context":
        context = super().get_context(context)
        context["environment"] = Environment(name=context["environment_name"], project=self.project, vars_={})

        self.project.provider.update_config({"region": context["aws_region"]})
        return context



import argparse
import os
from typing import Any

from strong_opx.codegen.project_generator import ProjectGenerator
from strong_opx.config import PROJECT_CONFIG_FILE, system_config
from strong_opx.exceptions import CommandError
from strong_opx.management.command import BaseCommand
from strong_opx.management.utils import select_environment, select_project, validate_project_name
from strong_opx.project import Project


class Command(BaseCommand):
    help_text = "Manage Strong OpX projects"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        subparsers = parser.add_subparsers(title="operation", description="Operation to execute")

        list_ = subparsers.add_parser("list", help="List all registered projects")
        list_.set_defaults(operation="list")

        register = subparsers.add_parser("register", help="Register an existing project")
        register.add_argument("path", help="Root of project")
        register.set_defaults(operation="register")

        unregister = subparsers.add_parser("unregister", help="Remove a registered project")
        unregister.add_argument("project_name", help="Name of project e.g. my-project", type=validate_project_name)
        unregister.set_defaults(operation="unregister")

        create = subparsers.add_parser(
            "create",
            help="Create an empty Strong OpX project",
        )
        create.add_argument("project_name", help="Name of project e.g. my-project")
        create.add_argument("--path", help="Location where to create project, defaults to current directory")
        create.add_argument("--aws-region", help="AWS Region Name. Defaults to us-east-1", default="us-east-1")
        create.set_defaults(operation="create")

        init = subparsers.add_parser(
            "init",
            help="Initialize Strong OpX project. This will create S3 bucket required for deployment.",
        )
        init.add_argument("--project", type=validate_project_name, help="Name of project e.g. my-project")
        init.set_defaults(operation="init")

        init = subparsers.add_parser(
            "init-env",
            help="Initialize Strong OpX Environment. This will create DynamoDB table required for terraform locks.",
        )
        init.add_argument("--project", type=validate_project_name, help="Name of project e.g. my-project")
        init.add_argument("--env", help="Name of Environment e.g. production")
        init.set_defaults(operation="init-env")

    def handle(self, operation=None, **options: Any):
        if operation is None:
            raise CommandError("Specify operation to execute. See --help for more info")

        if operation == "register":
            path = os.path.abspath(options["path"])
            project = Project.from_config(os.path.join(path, PROJECT_CONFIG_FILE))
            system_config.register_project(project.name, path)

        elif operation == "unregister":
            system_config.unregister_project(options["project_name"])

        elif operation == "list":
            for project in system_config.registered_projects:
                print(project, system_config.get_project_path(project), sep="\t")

        elif operation == "create":
            name = options["project_name"]
            full_path = os.path.join(os.path.abspath(options.get("path") or "."), name)

            ProjectGenerator(name, full_path).generate()
            system_config.register_project(name, full_path)

        elif operation == "init":
            select_project(options.get("project")).init()

        elif operation == "init-env":
            project = select_project(options.get("project"))
            select_environment(project, options.get("env")).init()



import os
import shutil
from typing import Generator

from strong_opx.codegen.questions import Question
from strong_opx.project import Project
from strong_opx.template import Context, FileTemplate, Template
from strong_opx.utils.prompt import select_prompt

TEMPLATE_POSTFIX = "-tpl"
BASE_TEMPLATE_DIR = os.path.join(os.path.dirname(os.path.dirname(__file__)), "templates")


class CodeGenerator:
    questions: dict[str, Question] = {}

    def __init__(self, project: "Project" = None):
        self.project = project

    def get_output_dir(self, context: "Context") -> str:
        raise NotImplementedError()

    def generate(self) -> None:
        context = Context({"project": self.project})
        for name, question in self.questions.items():
            context[name] = question.from_stdin()

        output_dir = self.get_output_dir(context)
        self._generate(output_dir, context)

    def _generate(self, output_dir: str, context: Context) -> None:
        raise NotImplementedError()


class TemplateGenerator(CodeGenerator):
    template_dir: str

    def get_context(self, context: "Context") -> "Context":
        context["project"] = self.project
        return context

    def list_templates(self) -> list[str]:
        template_dir = os.path.join(BASE_TEMPLATE_DIR, self.template_dir)

        return [d for d in os.listdir(template_dir) if os.path.isdir(os.path.join(template_dir, d))]

    def _generate(self, output_dir: str, context: Context) -> None:
        all_templates = self.list_templates()
        if len(all_templates) == 1:
            template_name = all_templates[0]
        else:
            template_name = select_prompt("Select Template", all_templates)

        self.render_template_dir(template_name, output_dir, context)

    def render_template_dir(self, template_name: str, output_dir: str, context: Context) -> None:
        context = self.get_context(context)
        template_dir = os.path.join(BASE_TEMPLATE_DIR, self.template_dir, template_name)

        for template_file in self.iter_template_files(template_dir):
            target_path = os.path.join(output_dir, Template(template_file).render(context))

            if template_file.endswith(TEMPLATE_POSTFIX):
                template = FileTemplate(os.path.join(template_dir, template_file))
                template.render_to_file(target_path[: -len(TEMPLATE_POSTFIX)], context)
            else:
                os.makedirs(os.path.dirname(target_path), exist_ok=True)
                shutil.copy(
                    os.path.join(template_dir, template_file),
                    target_path,
                )

    @staticmethod
    def iter_template_files(base_path: str) -> Generator[str, None, None]:
        dirs = [base_path]

        while dirs:
            directory = dirs.pop()
            for filename in os.listdir(directory):
                full_path = os.path.join(directory, filename)
                if os.path.isdir(full_path):
                    dirs.append(full_path)
                else:
                    yield os.path.relpath(full_path, base_path)



import logging
from typing import TYPE_CHECKING

import boto3
from botocore.exceptions import ClientError
from pydantic.dataclasses import dataclass

from strong_opx.providers.secret_provider import SecretProvider
from strong_opx.template import ObjectTemplate

if TYPE_CHECKING:
    from strong_opx.project import Environment

logger = logging.getLogger(__name__)


@dataclass
class SSMSecretProvider(SecretProvider):
    secret_length: int = 24
    parameter: str = None

    def get_secret(self, environment: "Environment"):
        if self.parameter is None:
            parameter = f"/strong-opx/{environment.project.name}/{environment.name}/vault-secret"
        else:
            parameter = ObjectTemplate(environment.base_context).render(self.parameter)

        ssm = boto3.client("ssm")
        try:
            response = ssm.get_parameter(Name=parameter, WithDecryption=True)
            secret = response["Parameter"]["Value"]
        except ClientError as e:
            error_code = e.response.get("Error", {}).get("Code")
            if error_code != "ParameterNotFound":
                raise e

            logger.info(f"Unable to locate {parameter}. Generating new secret for encryption")
            secret = self.generate_secret()
            ssm.put_parameter(
                Name=parameter,
                Description=f"Strong-OpX Secret for {environment.project.name} ({environment.name})",
                Value=secret,
                Type="SecureString",
                Tags=[
                    {"Key": "Project", "Value": environment.project.name},
                    {"Key": "Environment", "Value": environment.name},
                ],
            )

        return secret


SECRET_PROVIDERS = {
    "aws_ssm": SSMSecretProvider,
}



from typing import Optional

from pydantic.dataclasses import dataclass


@dataclass
class AzureConfig:
    resource_group: Optional[str] = None
    subscription_id: Optional[str] = None
    tenant_id: Optional[str] = None

    def dict(self) -> dict[str, str]:
        d = {}
        if self.resource_group:
            d["AZURE_RESOURCE_GROUP"] = self.resource_group

        if self.subscription_id:
            d["AZURE_SUBSCRIPTION_ID"] = self.subscription_id

        if self.tenant_id:
            d["AZURE_TENANT_ID"] = self.tenant_id

        return d



import os
from typing import TYPE_CHECKING

from strong_opx.exceptions import ImproperlyConfiguredError
from strong_opx.providers.aws.iam import get_current_account_id

if TYPE_CHECKING:
    from strong_opx.template import Context

AWS_CREDENTIALS_ENVIRON_KEYS = [
    "AWS_PROFILE",
    "AWS_ACCESS_KEY_ID",
    "AWS_SECRET_ACCESS_KEY",
    "AWS_SESSION_TOKEN",
]


def import_and_clean_environ_hook(context: "Context"):
    allow_credentials = context.get("AWS_PROFILE") is None
    context["ACCOUNT_ID"] = get_current_account_id

    for k in os.environ:
        if not k.startswith("AWS_"):
            continue

        if allow_credentials and k in AWS_CREDENTIALS_ENVIRON_KEYS:
            context[k] = os.environ[k]
        else:
            del os.environ[k]


def update_environ_hook(context: "Context"):
    for k in context:
        if k.startswith("AWS_"):
            os.environ[k] = context[k]

    if "AWS_REGION" not in context:
        raise ImproperlyConfiguredError(
            "AWS Region is not configured. Either configure that in project or environment under aws.region"
        )

    os.environ["AWS_DEFAULT_REGION"] = context["AWS_REGION"]



import os.path
from subprocess import CompletedProcess
from unittest import TestCase, mock

from pydantic import ValidationError

from strong_opx.exceptions import HelmError
from strong_opx.helm import HelmChart, HelmChartQualifiedName, HelmConfig, HelmManager
from tests.mocks import create_mock_environment, create_mock_kubernetes_platform, create_mock_project


class HelmChartTests(TestCase):
    def test_qualified_name(self):
        chart = HelmChart(repo="some-repo", chart="chart-a")
        self.assertEqual(chart.qualified_name, "default:chart-a")

    def test_qualified_name_custom_namespace(self):
        chart = HelmChart(repo="some-repo", chart="chart-b", namespace="some-ns")
        self.assertEqual(chart.qualified_name, "some-ns:chart-b")

    def test_name_equal_chart_if_missing(self):
        chart = HelmChart(repo="some-repo", chart="chart-c")
        self.assertEqual(chart.name, "chart-c")

    def test_specified_name(self):
        chart = HelmChart(repo="some-repo", chart="chart-c", name="chart-c-name")
        self.assertEqual(chart.name, "chart-c-name")


class HelmConfigTests(TestCase):
    def test_default_repos(self):
        config = HelmConfig()
        self.assertDictEqual(config.repos, {"stable": "https://charts.helm.sh/stable"})

    def test_default_repo_if_not_specified(self):
        config = HelmConfig(repos={"some-repo": "https://charts.example.com"})
        self.assertDictEqual(
            config.repos, {"some-repo": "https://charts.example.com/", "stable": "https://charts.helm.sh/stable"}
        )

    def test_invalid_repo_url(self):
        with self.assertRaises(ValidationError):
            HelmConfig(repos={"some-repo": "charts.example.com"})


class HelmManagerTests(TestCase):
    helm_run: mock.Mock

    def create_helm_manager(
        self, helm_config: HelmConfig, installed_helm_charts: list[HelmChartQualifiedName] = None
    ) -> HelmManager:
        project = create_mock_project(helm_config=helm_config)
        environment = create_mock_environment(project=project)
        platform = create_mock_kubernetes_platform(project=project, environment=environment)

        manager = HelmManager(platform)
        if installed_helm_charts is not None:
            manager.__dict__["installed_helm_charts"] = installed_helm_charts
        self.helm_run = manager.run = mock.MagicMock()
        return manager

    def test_installed_helm_charts(self):
        helm_manager = self.create_helm_manager(HelmConfig(charts=[]))

        self.helm_run.return_value = mock.create_autospec(
            spec=CompletedProcess,
            instance=True,
            stdout=b'[{"namespace": "default", "name": "milena"},{"namespace": "keda", "name": "ivan"}]',
        )

        self.assertListEqual(
            helm_manager.installed_helm_charts,
            [
                HelmChartQualifiedName("default", "milena"),
                HelmChartQualifiedName("keda", "ivan"),
            ],
        )

    def test_install(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1")]), installed_helm_charts=[]
        )

        helm_manager.apply(charts=[], upgrade=False)
        self.helm_run.assert_called_once_with(
            "install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
        )

    def test_install_oci_repo(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[HelmChart(repo="fake-oci-repo", chart="chart-1")],
                repos={"fake-oci-repo": "oci://fake-oci-repo"},
            ),
            installed_helm_charts=[],
        )

        helm_manager.apply(charts=[], upgrade=False)
        self.helm_run.assert_called_once_with(
            "install",
            "chart-1",
            "oci://fake-oci-repo",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
        )

    def test_install_custom_timeout(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1", timeout="20 years")]), installed_helm_charts=[]
        )

        helm_manager.apply(charts=[], upgrade=False)
        self.helm_run.assert_called_once_with(
            "install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "20 years",
        )

    def test_install__additional_args(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1")]), installed_helm_charts=[]
        )
        helm_manager.apply(upgrade=False, additional_args=("--debug",))
        self.helm_run.assert_called_once_with(
            "install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
            "--debug",
        )

    def test_install__specific_version(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1", version="0.1")]), installed_helm_charts=[]
        )
        helm_manager.apply(upgrade=False)
        self.helm_run.assert_called_once_with(
            "install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
            "--version",
            "0.1",
        )

    def test_install__specific_chart(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[
                    HelmChart(repo="stable", chart="chart-1"),
                    HelmChart(repo="stable", chart="chart-2"),
                ]
            ),
            installed_helm_charts=[],
        )
        helm_manager.apply(charts=["chart-1"], upgrade=False)
        self.helm_run.assert_called_once_with(
            "install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
        )

    def test_install__skip_already_installed(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[
                    HelmChart(repo="stable", chart="chart-1"),
                    HelmChart(repo="stable", chart="chart-2"),
                ]
            ),
            installed_helm_charts=[HelmChartQualifiedName("default", "chart-1")],
        )
        helm_manager.apply(upgrade=False)
        self.helm_run.assert_called_once_with(
            "install",
            "chart-2",
            "chart-2",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
        )

    def test_install__missing_custom_values(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1", values="helm/custom-values.yml")]),
            installed_helm_charts=[],
        )

        with self.assertRaises(HelmError):
            helm_manager.apply(upgrade=False)

        self.helm_run.assert_not_called()

    def test_upgrade(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1")]), installed_helm_charts=[]
        )
        helm_manager.apply(upgrade=True)
        self.helm_run.assert_called_once_with(
            "upgrade",
            "--install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
        )

    def test_upgrade__specific_version(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1", version="0.1")]), installed_helm_charts=[]
        )
        helm_manager.apply(upgrade=True)
        self.helm_run.assert_called_once_with(
            "upgrade",
            "--install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
            "--version",
            "0.1",
        )

    def test_upgrade__specific_chart(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[
                    HelmChart(repo="stable", chart="chart-1"),
                    HelmChart(repo="stable", chart="chart-2"),
                ]
            ),
            installed_helm_charts=[],
        )
        helm_manager.apply(charts=["chart-1"], upgrade=True)
        self.helm_run.assert_called_once_with(
            "upgrade",
            "--install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
        )

    @mock.patch("strong_opx.helm.os.path.exists")
    @mock.patch("strong_opx.helm.FileTemplate", new=mock.MagicMock())
    @mock.patch("strong_opx.helm.tempfile.NamedTemporaryFile")
    def test_install_custom_values(self, named_temporary_file_mock: mock.Mock, path_exists_mock: mock.Mock):
        path_exists_mock.return_value = True
        named_temporary_file_mock.return_value.__enter__.return_value.name = "values-file.yml"

        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1", values="helm/chart.yml")]),
            installed_helm_charts=[],
        )

        helm_manager.apply(upgrade=False)
        self.helm_run.assert_called_once_with(
            "install",
            "chart-1",
            "chart-1",
            "--repo",
            "https://charts.helm.sh/stable",
            "-n",
            "default",
            "--create-namespace",
            "--wait",
            "--timeout",
            "5m0s",
            "--values",
            "values-file.yml",
        )

    @mock.patch("strong_opx.helm.os.path.exists")
    @mock.patch("strong_opx.helm.FileTemplate")
    @mock.patch("strong_opx.helm.tempfile.NamedTemporaryFile")
    def test_install_custom_values_rendering(
        self, named_temporary_file_mock: mock.Mock, file_template_mock: mock.Mock, path_exists_mock: mock.Mock
    ):
        path_exists_mock.return_value = True
        named_temporary_file_mock.return_value.__enter__.return_value.name = "values-tmp-file.yml"

        helm_manager = self.create_helm_manager(
            HelmConfig(charts=[HelmChart(repo="stable", chart="chart-1", values="helm/chart.yml")]),
            installed_helm_charts=[],
        )

        helm_manager.apply(upgrade=False)
        file_template_mock.assert_called_once_with(os.path.join(helm_manager.platform.project.path, "helm/chart.yml"))
        file_template_mock.return_value.render_to_file.assert_called_once_with(
            "values-tmp-file.yml", helm_manager.platform.environment.context
        )

    def test_upgrade__upgrade_already_installed(self):
        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[
                    HelmChart(repo="stable", chart="chart-1"),
                    HelmChart(repo="stable", chart="chart-2"),
                ]
            ),
            installed_helm_charts=[HelmChartQualifiedName("default", "chart-1")],
        )
        helm_manager.apply(upgrade=True)
        self.helm_run.assert_has_calls(
            [
                mock.call(
                    "upgrade",
                    "--install",
                    "chart-1",
                    "chart-1",
                    "--repo",
                    "https://charts.helm.sh/stable",
                    "-n",
                    "default",
                    "--create-namespace",
                    "--wait",
                    "--timeout",
                    "5m0s",
                ),
                mock.call(
                    "upgrade",
                    "--install",
                    "chart-2",
                    "chart-2",
                    "--repo",
                    "https://charts.helm.sh/stable",
                    "-n",
                    "default",
                    "--create-namespace",
                    "--wait",
                    "--timeout",
                    "5m0s",
                ),
            ],
            any_order=True,
        )

    @mock.patch("strong_opx.helm.input_boolean")
    def test_prune__nothing(self, input_boolean_mock: mock.Mock):
        input_boolean_mock.return_value = True

        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[
                    HelmChart(repo="stable", chart="chart-1"),
                    HelmChart(repo="stable", chart="chart-2"),
                ]
            ),
            installed_helm_charts=[HelmChartQualifiedName("default", "chart-1")],
        )

        helm_manager.prune()
        self.helm_run.assert_not_called()

    @mock.patch("strong_opx.helm.input_boolean")
    def test_prune__chart(self, input_boolean_mock: mock.Mock):
        input_boolean_mock.return_value = True
        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[
                    HelmChart(repo="stable", chart="chart-1"),
                    HelmChart(repo="stable", chart="chart-2"),
                ]
            ),
            installed_helm_charts=[
                HelmChartQualifiedName("default", "chart-2"),
                HelmChartQualifiedName("default", "chart-3"),
            ],
        )
        helm_manager.prune()
        self.helm_run.assert_called_once_with("uninstall", "chart-3", "-n", "default")

    @mock.patch("strong_opx.helm.input_boolean")
    def test_prune__additional_args(self, input_boolean_mock: mock.Mock):
        input_boolean_mock.return_value = True
        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[
                    HelmChart(repo="stable", chart="chart-1"),
                    HelmChart(repo="stable", chart="chart-2"),
                ]
            ),
            installed_helm_charts=[
                HelmChartQualifiedName("default", "chart-2"),
                HelmChartQualifiedName("default", "chart-3"),
            ],
        )
        helm_manager.prune(additional_args=("--debug",))
        self.helm_run.assert_called_once_with("uninstall", "chart-3", "-n", "default", "--debug")

    @mock.patch("strong_opx.helm.input_boolean")
    def test_prune__reject_input(self, input_boolean_mock: mock.Mock):
        input_boolean_mock.return_value = False
        helm_manager = self.create_helm_manager(
            HelmConfig(
                charts=[
                    HelmChart(repo="stable", chart="chart-1"),
                    HelmChart(repo="stable", chart="chart-2"),
                ]
            ),
            installed_helm_charts=[
                HelmChartQualifiedName("default", "chart-2"),
                HelmChartQualifiedName("default", "chart-3"),
            ],
        )
        helm_manager.prune()
        self.helm_run.assert_not_called()



import logging
import re
from typing import TYPE_CHECKING, Generator, Optional

from strong_opx.exceptions import RepositoryNotFoundException
from strong_opx.providers.discovery import current_provider_name
from strong_opx.template import Template
from strong_opx.utils.module_loading import import_module_attr_if_exists

if TYPE_CHECKING:
    from strong_opx.project.environment import Environment

logger = logging.getLogger(__name__)
DEFAULT_DOCKER_TAG = "latest"


class AbstractDockerRegistry:
    max_images_to_keep = 8
    revision_tag_re = re.compile(r"^(.+)-([0-9]+)(\.[a-z0-9]+)?$")

    def __init__(self, environment: "Environment"):
        self.environment = environment

    def login(self):
        raise NotImplementedError()

    def create_repository(self, repository_name: str) -> str:
        raise NotImplementedError()

    def get_repository_uri(self, repository_name: str) -> Optional[str]:
        raise NotImplementedError()

    def get_or_create_repository_uri(self, repository_name: str) -> str:
        try:
            return self.get_repository_uri(repository_name)
        except RepositoryNotFoundException:
            return self.create_repository(repository_name)

    def revision_from_tag(self, tag: str) -> int:
        match = self.revision_tag_re.match(tag)
        if match:
            env = match.group(1)
            if env == self.environment.name:
                return int(match.group(2))

        return 0

    def tag_from_revision(self, revision: int) -> str:
        tag = f"{self.environment.name}-{revision}"
        version_hash = self.environment.project.git_revision_hash()

        if version_hash:
            tag = f"{tag}.{version_hash}"

        return tag

    def iter_image_tags(self, repository_name: str) -> Generator[str, None, None]:
        raise NotImplementedError()

    def get_latest_revision(self, repository_name: str) -> int:
        revision = 0

        for tag in self.iter_image_tags(repository_name):
            revision = max(self.revision_from_tag(tag), revision)

        return revision

    def get_latest_image_uri(self, repository_name: str, render_repository_name: bool = False) -> str:
        if render_repository_name:
            repository_name = Template(repository_name).render(self.environment.context)

        revision = 0
        latest_tag = DEFAULT_DOCKER_TAG

        for tag in self.iter_image_tags(repository_name):
            tag_revision = self.revision_from_tag(tag)
            if tag_revision > revision:
                revision = tag_revision
                latest_tag = tag

        try:
            repository_uri = self.get_repository_uri(repository_name)
        except RepositoryNotFoundException:
            repository_uri = repository_name

        return f"{repository_uri}:{latest_tag}"


def current_docker_registry(environment: "Environment") -> Optional[AbstractDockerRegistry]:
    registry_class: Optional[type[AbstractDockerRegistry]] = import_module_attr_if_exists(
        f"strong_opx.providers.{current_provider_name()}.docker_registry",
        "DockerRegistry",
    )
    if registry_class:
        return registry_class(environment)



from functools import cached_property
from typing import TYPE_CHECKING, Callable

from strong_opx.providers import Provider
from strong_opx.providers.gcloud.context_hooks import update_environ_hook
from strong_opx.providers.gcloud.credentials import GCloudConfig

if TYPE_CHECKING:
    from strong_opx.template import Context


class GCloudProvider(Provider):
    config: GCloudConfig
    compute_instance_id_re = r"^i-[0-9a-f]+$"

    @cached_property
    def gcp_project_path(self):
        return f"projects/{self.config.project}/locations/{self.config.compute_region}"

    def get_additional_context_hooks(self) -> tuple[Callable[["Context"], None], ...]:
        return (
            self.update_context,
            update_environ_hook,
        )

    def update_context(self, context: "Context"):
        context.update(self.config.dict())



import boto3

from strong_opx.providers.compute import ComputeInstanceDescription, ComputeInstanceState
from strong_opx.utils.mapping import CaseInsensitiveMultiTagDict


def transform_instance(instance: dict) -> ComputeInstanceDescription:
    tags = CaseInsensitiveMultiTagDict()
    for tag in instance.get("Tags", []):
        tags[tag["Key"]] = tag["Value"]

    try:
        state = ComputeInstanceState(instance["State"]["Name"])
    except ValueError:
        state = ComputeInstanceState.UNKNOWN

    return ComputeInstanceDescription(
        instance_id=instance["InstanceId"],
        state=state,
        public_ip=instance.get("PublicIpAddress"),
        private_ip=instance.get("PrivateIpAddress"),
        tags=tags,
    )


def describe_instances(**kwargs) -> list[ComputeInstanceDescription]:
    response = boto3.client("ec2").describe_instances(**kwargs)

    instances = []
    for reservation in response["Reservations"]:
        for instance in reservation.get("Instances", []):
            if instance["State"]["Name"] == "terminated":
                continue  # Skip terminated instances

            instances.append(transform_instance(instance))

    return instances



from strong_opx.providers.compute import ComputeInstance, ComputeInstanceDescription
from strong_opx.providers.discovery import current_provider_class, current_provider_error_handler, select_provider
from strong_opx.providers.docker_registry import current_docker_registry
from strong_opx.providers.provider import Provider
from strong_opx.providers.secret_provider import SecretProvider



from collections import UserDict
from typing import Any, Callable, Generator, MutableMapping

from pydantic_core import CoreSchema, core_schema

NOT_SPECIFIED = object()


class LazyValue:
    def __init__(self, resolver: Callable[[], Any]):
        self.resolver = resolver

    def resolve(self) -> Any:
        return self.resolver()

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.resolver}>"


class LazyDict(MutableMapping):
    def __init__(self, *args, **kwargs):
        self._data: dict[str, Any] = dict(*args, **kwargs)

    def __missing__(self, key):
        raise KeyError(key)

    def __getitem__(self, key: str) -> Any:
        """
        Retrieves the value associated with 'key'. If the value is a `LazyValue`, that will be resolved and
        stored in the object for future use.

        @param key: Key to get value of
        @return: The value corresponding to the key
        @raise KeyError if the specified key does not exist
        """
        value = self.get(key, default=NOT_SPECIFIED)
        if value is NOT_SPECIFIED:
            return self.__missing__(key)

        return value

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key):
        del self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Generator[str, None, None]:
        yield from self._data

    def __repr__(self) -> str:
        r = ", ".join(f"{k!r}: {self.get(k, resolve=False)!r}" for k in self)
        return f"{self.__class__.__name__}({{{r}}})"

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def set_lazy(self, key: str, resolver: Callable[[], Any]) -> None:
        """
        Set a lazy value for the specified key. The value will be resolved when it is first accessed.
        `resolver` will only be called once and the result will be cached.

        :param key: Key to set value for
        :param resolver: Function to resolve the value
        """
        self._data[key] = LazyValue(resolver)

    def get(self, key: str, default=None, resolve: bool = True) -> Any:
        """
        Retrieves the value associated with `key`. Optionally if the value is a `LazyValue`,
        that will be resolved and the results are stored and returned from get().

        @param key: Key to get value of
        @param default: Value to return if the key does not exist
        @param resolve: Whether to resolve a `LazyValue` if found
        @return: The value corresponding to the key
        """
        value = self._data.get(key, default)
        if resolve and isinstance(value, LazyValue):
            value = value.resolve()
            self._data[key] = value

        return value


class CaseInsensitiveMultiTagDict(UserDict[str, list[str]]):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.update(dict(*args, **kwargs))

    def __setitem__(self, key: str, value: str) -> None:
        key = key.lower()
        if key in self.data:
            if value not in self.data[key]:
                self.data[key].append(value)
        else:
            self.data[key] = [value]

    def __getitem__(self, key: str) -> list[str]:
        return self.data[key.lower()]

    def __contains__(self, key: str) -> bool:
        return key.lower() in self.data

    def __delitem__(self, key: str) -> None:
        key = key.lower()
        if key in self.data:
            del self.data[key]
        else:
            raise KeyError(key)

    def get(self, key: str, default=NOT_SPECIFIED) -> list[str]:
        if default is NOT_SPECIFIED:
            default = []

        return self.data.get(key.lower(), default)

    @staticmethod
    def __get_pydantic_core_schema__(source_type, handler) -> CoreSchema:
        return core_schema.dict_schema(
            keys_schema=core_schema.str_schema(),
            values_schema=core_schema.union_schema(
                [core_schema.str_schema(), core_schema.list_schema(core_schema.str_schema())]
            ),
        )



import os

import jinja2

from strong_opx.config import opx_config
from strong_opx.template.context import Context
from strong_opx.template.template import Template
from strong_opx.utils.tracking import OpxString, Position, set_position


class FileTemplate:
    def __init__(self, file_path: str):
        self.file_path = file_path

    @property
    def content(self) -> str:
        with open(self.file_path) as f:
            return f.read()

    def render(self, context: Context) -> str:
        if opx_config.templating_engine == "jinja2":
            return self._render_with_jinja2(context)

        return self._default_renderer(context)

    def _default_renderer(self, context: Context) -> str:
        content = OpxString(self.content)
        set_position(content, self.file_path, Position(1, 1), None)
        return Template(content).render(context)

    def _render_with_jinja2(self, context: Context) -> str:
        loader = jinja2.FileSystemLoader(os.path.dirname(self.file_path))
        environment = jinja2.Environment(loader=loader)
        template = environment.from_string(self.content)
        return template.render(**context.as_dict())

    def render_to_file(self, target_path: str, context: Context) -> None:
        os.makedirs(os.path.dirname(target_path), exist_ok=True)
        with open(target_path, "w") as f:
            f.write(self.render(context))



from typing import Optional

from pydantic import ValidationError

from strong_opx.exceptions import ErrorDetail, YAMLError
from strong_opx.utils.tracking import Position, get_position

ERROR_MESSAGE_BY_TYPE = {
    "missing": "Missing required field",
}


def get_position_by_path(
    input_values: dict, path: tuple
) -> tuple[Optional[str], Optional[Position], Optional[Position]]:
    parent = None
    current = input_values
    for part in path:
        parent = current
        if isinstance(current, dict) and part in current:
            current = current[part]
        else:
            return None, None, None

    position = get_position(current)
    if position[0] is None:
        actual_key = next((k for k in parent.keys() if k == path[-1]), None)
        return get_position(actual_key)

    return position


def translate_pydantic_errors(input_values: dict, ex: ValidationError) -> Exception:
    translated = []

    for error in ex.errors():
        file_name = get_position(input_values)[0]
        _, start_offset, end_offset = get_position_by_path(input_values, error["loc"])
        message = ERROR_MESSAGE_BY_TYPE.get(error["type"], error["msg"])

        if start_offset is None:
            loc = " > ".join(error["loc"])
            message = f"{loc}: {message}"

        translated.append(
            ErrorDetail(
                error=message,
                file_path=file_name,
                start_pos=start_offset,
                end_pos=end_offset,
            )
        )

    return YAMLError(translated)



import argparse
from typing import Any

from strong_opx.config import system_config
from strong_opx.management.command import BaseCommand
from strong_opx.management.utils import validate_project_name


class Command(BaseCommand):
    help_text = "Get or change Strong-OpX configuration"

    def add_arguments(self, parser: argparse.ArgumentParser):
        parser.add_argument("--remove", action="store_true", default=False, help="Remove configuration instead")

        parser.add_argument(
            "--project",
            dest="project_name",
            type=validate_project_name,
            help="Project name if this config is project specific. If not specified, applies to all projects "
            "(unless overridden by project specific config).",
        )

        parser.add_argument("key", help="Configuration key")
        parser.add_argument("value", nargs="?", help="Configuration value")

    def handle(self, key: str, project_name: str = None, value=None, **options: Any):
        if project_name is None:
            config = system_config
        else:
            config = system_config.get_project_config(project_name)

        parts = key.split(".", 1)
        if len(parts) == 1:
            section = "default"
        else:
            section, key = parts

        if section not in config:
            config.add_section(section)

        if options["remove"]:
            config.remove_option(section, key)
        elif value is None:
            value = config.get(section, key, fallback=None)
            if value is not None:
                print(value)

            return
        else:
            config.set(section, key, value)

        config.save()



import os
from typing import Union

from botocore.exceptions import ClientError, NoCredentialsError
from colorama import Style

from strong_opx.exceptions import CommandError, RepositoryNotFoundException
from strong_opx.project import Project
from strong_opx.providers.aws.credentials import AWSCredentialConfig


class CredentialError(CommandError):
    def __str__(self):
        project = Project.current()
        profile_name = project.config.get("aws", "aws_profile", fallback=None)

        if profile_name is None:
            message = (
                f"No AWS profile is not configured for current project ({project.name})."
                "\n\n"
                "Configure it using the following command:"
                "\n\n"
                f"{Style.BRIGHT}    strong-opx config aws.aws_profile <aws-profile-name-here> --project "
                f"{project.name}{Style.RESET_ALL}\n"
            )
        else:
            message = (
                f'No credentials found for profile "{profile_name}"'
                "\n\n"
                f"To configured AWS credentials for profile {profile_name}, run the following command:"
                "\n\n"
                f"{Style.BRIGHT}    strong-opx aws:configure --profile {profile_name}{Style.RESET_ALL}\n"
            )

        return message


class CredentialExpiredError(CommandError):
    def __str__(self):
        profile_name = os.getenv("AWS_PROFILE", "default")

        aws_config = AWSCredentialConfig()
        if not aws_config.are_credential_expired(profile_name):
            return super().__str__()

        return (
            f'Credentials for profile "{profile_name}" have expired'
            f"\n\n"
            "MFA is enabled for this profile. Run following command to get temporary credentials:"
            "\n\n"
            f"{Style.BRIGHT}    strong-opx aws:mfa --profile {profile_name} <mfa-token-code>{Style.RESET_ALL}\n"
        )


AWS_ERROR_CODE_TO_EXCEPTION = {
    "ExpiredToken": CredentialExpiredError,
    "RequestExpired": CredentialExpiredError,
    "ExpiredTokenException": CredentialExpiredError,
    "RepositoryNotFoundException": RepositoryNotFoundException,
}


def handle_boto_error(ex: Union[ClientError, NoCredentialsError], ignore: tuple[str, ...] = ()) -> None:
    if isinstance(ex, NoCredentialsError):
        raise CredentialError()

    error = (ex.response or {}).get("Error", {})
    code = error.get("Code")
    if code in ignore:
        return  # Ignore this exception

    constructor = AWS_ERROR_CODE_TO_EXCEPTION.get(code)
    if not constructor:
        raise

    message = error.get("Message", str(ex))
    raise constructor(message)



import logging
import os
from functools import cached_property
from typing import TYPE_CHECKING, Annotated, Any, Callable, TypeVar

from pydantic import Field, ValidationError, create_model

from strong_opx import yaml
from strong_opx.exceptions import ProjectEnvironmentError
from strong_opx.platforms import ALL_PLATFORMS, TPlatform
from strong_opx.project.context_hooks import EnvironHook, ProjectContextHook
from strong_opx.providers import current_docker_registry
from strong_opx.template import Context
from strong_opx.utils.validation import translate_pydantic_errors

if TYPE_CHECKING:
    from strong_opx.platforms.deployments import DeploymentProvider
    from strong_opx.project import Project

T = TypeVar("T")
logger = logging.getLogger(__name__)


class Environment:
    def __init__(self, name: str, project: "Project", vars_: dict[str, str]):
        self.name = name
        self.vars = vars_

        self.project = project
        self.platforms: list[TPlatform] = []
        self.path = os.path.join(project.environments_dir, self.name)
        self.base_context = self.create_context()

    def register_platform(self, platform: TPlatform):
        self.platforms.append(platform)
        platform.init_context(self.base_context)

    def select_platform(self, platform_cls: type[TPlatform]) -> TPlatform:
        for platform in self.platforms:
            if isinstance(platform, platform_cls):
                return platform

        raise ProjectEnvironmentError(f"Environment {self.name} does not support {platform_cls.__name__}")

    def get_context_hooks(self) -> tuple[Callable[["Context"], None], ...]:
        return (
            ProjectContextHook(self.project),
            EnvironHook("STRONG_OPX_"),
            *self.project.provider.get_additional_context_hooks(),
        )

    def create_context(self) -> Context:
        context = Context(self.vars)
        context["ENVIRONMENT"] = self.name
        context["DOCKER_REGISTRY"] = current_docker_registry(self)

        for hook in self.get_context_hooks():
            hook(context)

        return context

    @cached_property
    def vault_secret(self):
        return self.project.secret_provider.get_secret(self)

    @cached_property
    def context(self) -> Context:
        context = self.base_context.chain()

        var_paths = self.project.vars_config.get_paths(self)
        for file_path in var_paths:
            abs_path = os.path.join(self.project.path, file_path)
            if os.path.exists(abs_path):
                logger.debug(f"Loading vars from {file_path}")
                context.load_from_file(abs_path)
            else:
                logger.warning(f"Unable to locate {file_path}")

        return context

    def init(self):
        self.project.provider.init_environment(self)

    def deploy(self, nodes: list[str]) -> None:
        nodes_by_provider = {}
        for platform in self.platforms:
            for provider_class in platform.deployment_providers:
                provider: "DeploymentProvider" = provider_class(
                    project=self.project, environment=self, platform=platform
                )
                selected_nodes = provider.select_nodes(nodes)
                if selected_nodes:
                    nodes_by_provider[provider] = selected_nodes

        if not nodes_by_provider:
            logger.error("Nothing to deploy")

        for provider, nodes in nodes_by_provider.items():
            for node in nodes:
                logger.info(f"Deploying {node} using {provider.name}")
                provider.deploy_node(node)


def load_environment(environment_name: str, project: "Project") -> Environment:
    config_path = os.path.join(project.environments_dir, environment_name, "config.yml")
    if not os.path.exists(config_path) or not os.path.isfile(config_path):
        raise ProjectEnvironmentError(f"{environment_name} is not a valid strong-opx environment")

    config = yaml.load(config_path)

    environment_config_cls = create_model(
        "EnvironmentConfig",
        vars=(dict[str, Any], Field(default_factory=dict)),
        additional_provider_config=Annotated[dict[str, Any], Field(alias=project.provider.name, default_factory=dict)],
    )

    try:
        validated_config = environment_config_cls(**config)
    except ValidationError as ex:
        raise translate_pydantic_errors(config, ex)

    project.provider.update_config(validated_config.additional_provider_config)
    environment = Environment(name=environment_name, project=project, vars_=validated_config.vars)

    for platform_cls in ALL_PLATFORMS:
        platform = platform_cls.from_config(project=project, environment=environment, **config)
        if platform is not None:
            environment.register_platform(platform)

    return environment



import logging
import os
from functools import cached_property
from typing import Optional

from strong_opx.config import PROJECT_CONFIG_FILE, opx_config, system_config
from strong_opx.config.hierarchical import HierarchicalConfig
from strong_opx.exceptions import ImproperlyConfiguredError, ProcessError, ProjectEnvironmentError, ProjectError
from strong_opx.helm import HelmConfig
from strong_opx.project.config import ProjectConfig
from strong_opx.project.environment import Environment, load_environment
from strong_opx.project.vars import VariableConfig
from strong_opx.providers import Provider
from strong_opx.providers.secret_provider import SecretProvider
from strong_opx.utils.shell import shell

logger = logging.getLogger(__name__)


class Project:
    selected_environment: Optional["Environment"] = None

    def __init__(
        self,
        name: str,
        path: str,
        provider: Provider,
        secret_provider: SecretProvider,
        vars_config: VariableConfig,
        helm_config: HelmConfig,
    ):
        self.provider = provider
        self.name = name
        self.path = path
        self.helm_config = helm_config

        self.secret_provider = secret_provider
        self.vars_config = vars_config
        self.config = HierarchicalConfig([system_config.get_project_config(self.name), system_config])
        self.environments_dir = os.path.join(self.path, "environments")

    def __new__(cls, *args, **kwargs):
        instance = getattr(cls, "_instance", None)
        if instance is None:
            instance = super().__new__(cls)
            setattr(cls, "_instance", instance)

        return instance

    @classmethod
    def current(cls) -> Optional["Project"]:
        return getattr(cls, "_instance", None)

    @cached_property
    def environments(self) -> list[str]:
        if not os.path.exists(self.environments_dir):
            return []

        return [
            filename
            for filename in os.listdir(self.environments_dir)
            if os.path.isdir(os.path.join(self.environments_dir, filename))
        ]

    def select_environment(self, name: str) -> Environment:
        if name not in self.environments:
            raise ProjectEnvironmentError(f"Unknown environment: {name}")

        self.selected_environment = load_environment(environment_name=name, project=self)
        return self.selected_environment

    def git_revision_hash(self) -> Optional[str]:
        try:
            return (
                shell(
                    ["git", "rev-parse", "--short", "HEAD"],
                    capture_output=True,
                    cwd=self.path,
                )
                .stdout.decode("utf8")
                .strip()
            )
        except ProcessError:
            pass

    def init(self) -> None:
        self.provider.init_project(self)

    @classmethod
    def from_name(cls, project_name: str) -> "Project":
        project_path = system_config.get_project_path(project_name)
        config_path = os.path.join(project_path, PROJECT_CONFIG_FILE)

        if not os.path.exists(config_path) or not os.path.isfile(config_path):
            raise ProjectError(f"{project_name} is not a valid strong-opx project")

        project = cls.from_config(config_path)
        if project.name != project_name:
            raise ImproperlyConfiguredError(
                "Project name specified in config does not match the locally registered name"
            )

        return project

    @classmethod
    def from_config(cls, config_path: str) -> "Project":
        config = ProjectConfig.from_file(config_path)
        if config.strong_opx:
            opx_config.update(config.strong_opx)

        project_path = os.path.dirname(config_path)
        if config.dirname:
            project_path = os.path.join(project_path, config.dirname)

        return Project(
            name=config.name,
            path=project_path,
            secret_provider=config.secret,
            vars_config=config.vars,
            helm_config=config.helm,
            provider=config.provider,
        )



from typing import TYPE_CHECKING

from azure.core.exceptions import ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.keyvault.secrets import SecretClient
from pydantic.dataclasses import dataclass

from strong_opx.providers.secret_provider import SecretProvider
from strong_opx.template import ObjectTemplate

if TYPE_CHECKING:
    from strong_opx.project import Environment


@dataclass
class AzureKeyVaultSecretProvider(SecretProvider):
    keyvault_url: str
    parameter: str = None
    secret_length: int = 24

    def get_secret(self, environment: "Environment"):
        if self.parameter is None:
            parameter = f"/strong-opx/{environment.project.name}/{environment.name}/vault-secret"
        else:
            parameter = ObjectTemplate(environment.base_context).render(self.parameter)

        keyvault_url = ObjectTemplate(environment.base_context).render(self.keyvault_url)

        credential = DefaultAzureCredential()
        client = SecretClient(vault_url=keyvault_url, credential=credential)

        try:
            secret = client.get_secret(parameter).value
        except ResourceNotFoundError:
            secret = self.generate_secret()
            client.set_secret(
                name=parameter,
                value=secret,
                tags={"Project": environment.project.name, "Environment": environment.name},
            )

        return secret


SECRET_PROVIDERS = {
    "keyvault": AzureKeyVaultSecretProvider,
}



import ast
import functools
import re
from typing import Union

from strong_opx.exceptions import TemplateError
from strong_opx.template.registry import TEMPLATE_FILTERS
from strong_opx.template.variable import VariableRef, VariableStore
from strong_opx.utils.tracking import Position, get_position

CTX_LOAD = ast.Load()
CTX_STORE = ast.Store()
VAR_NAME_RE = re.compile(r"^[a-zA-Z_][a-zA-Z_0-9]*$")

OUTPUT_VAR_NAME = "_opx_out_"
CONTEXT_VAR_NAME = "_opx_ctx_"
INCLUDE_VAR_NAME = "_opx_include_"


class TemplateCompiler:
    """
    This class parses expressions that are passed to the class and builds a Python Abstract Syntax Tree (AST).

    An instance of the class is generally used like so:
    1. The caller has a collection of tokens it would like to parse
    2. The caller instantiates a TemplateCompiler
    3. For each token, the caller calls one of the TemplateCompiler::compile_*() methods
    4. The caller calls TemplateCompiler::finalize() to halt parsing and retrieve a built AST

    The caller may also make use of TemplateCompiler fields:
    * variables: A `VariableStore` objects that keeps track of all the variables that are defined/accessed in
                 the template.
    * global_vars: A set of variable names whose values are expected to be provided by the caller, when the AST is
                   transformed into code. The TemplateCompiler does NOT generate that code.

    Other fields are:
    * code_modules: code_modules[0] contains all the compiled expressions of self.value. It is what we return to a
                    caller with TemplateCompiler::finalize(). It is also used to keep track of blocks of code.
                    If len(self.code_modules) > 1, the caller is in the middle of parsing a block.
    * selected_body: The body of the currently selected code module. It is usually code_modules[-1].body, but may point
                     to other fields of code_modules[-1], depending on the block(s) being parsed.
    """

    def __init__(self, value: str, as_string: bool = True):
        self.value = value
        self.as_string = as_string

        self.variables = VariableStore()

        self.file_path, initial_pos, _ = get_position(value)
        if initial_pos:
            self.initial_line_no, self.initial_col_offset = initial_pos
        else:
            self.initial_line_no, self.initial_col_offset = 1, 1

        # LRU cache should be local to instance
        self.offset_to_position = functools.lru_cache(maxsize=32)(self.offset_to_position)
        self.code_modules: list[Union[ast.Module, ast.stmt]] = [ast.Module(body=[], type_ignores=[])]
        self.selected_body: list[ast.stmt] = self.code_modules[0].body

        if self.as_string:
            self.append = self.append_as_str

    def offset_to_position(self, offset: int) -> Position:
        """
        Return the line number and column number of a character, where it resides in its source string. It is NOT the
        line/column position of the character in self.value.

        This function treats self.value as an array of characters. The offset parameter is an index to that array.
        This function returns the line number and column number of that index, when splitting self.value into lines,
        with a new line starting at each newline (\n) character.

        The function takes the initial line number and column number into account to calculate the position in the
        source string.

        :param offset: 0-based index in template
        :return: [1-based index of line, 1-based index in that line]
        """
        return Position.from_offset(self.value, offset, self.initial_line_no, self.initial_col_offset)

    def set_location(self, node: ast.AST, start_offset: int, end_offset: int) -> None:
        """
        Given an Abstract Syntax Tree (AST) node with a 'lineno' attribute, update that node's start/end to the
        provided start/end

        :param node: Abstract Syntax Tree node to update
        :param start_offset: Start location to update AST to
        :param end_offset: End location to update AST to
        :return: (none)
        """
        if "lineno" in node._attributes:
            start_pos = self.offset_to_position(start_offset)
            end_pos = self.offset_to_position(end_offset)

            node.lineno = start_pos.line
            node.col_offset = start_pos.column

            node.end_lineno = end_pos.line
            node.end_col_offset = end_pos.column

    @staticmethod
    def str_strip(value: str, offset: int) -> tuple[str, int, int]:
        """
        Strips a string of leading and trailing whitespace, and returns the stripped string, along with the start and
        end offsets which bound the stripped string.

        @param value: The token to strip of whitespace
        @param offset: The offset of the start of the token
        @return: (stripped token, start offset of stripped token, end offset of stripped token)
        """
        t1 = value.lstrip()
        t1_len = len(t1)

        t2 = t1.rstrip()
        t2_len = len(t2)

        start_offset = offset + len(value) - t1_len
        end_offset = start_offset + t2_len

        return t2, start_offset, end_offset

    @staticmethod
    def split_on_whitespace(expr: str) -> tuple[str, str]:
        i = expr.find(" ")
        if i == -1:
            return expr, ""

        return expr[:i], expr[i:]

    def add_statement(self, value: ast.stmt, start_offset: int, end_offset: int) -> None:
        self.set_location(value, start_offset, end_offset)
        self.selected_body.append(value)

    def append(self, value, start_offset: int, end_offset: int) -> None:
        value: ast.Assign = ast.Assign(targets=[ast.Name(id=OUTPUT_VAR_NAME, ctx=CTX_STORE)], value=value)
        self.add_statement(value, start_offset, end_offset)

    def append_as_str(self, value, start_offset: int, end_offset: int) -> None:
        if not isinstance(value, ast.Constant):
            value: ast.Call = ast.Call(ast.Name(id="str", ctx=CTX_LOAD), args=[value], keywords=[])

        n_lines_append = ast.Attribute(value=ast.Name(id="lines", ctx=CTX_LOAD), attr="append", ctx=CTX_LOAD)
        value: ast.Expr = ast.Expr(value=ast.Call(n_lines_append, args=[value], keywords=[]))

        self.add_statement(value, start_offset, end_offset)

    def start_block(self, block: ast.stmt, start_offset: int, end_offset: int):
        """
        Used with ast.stmt subclasses which have "blocks" in Python. For example, 'if', 'for', 'while', 'try', etc.

        This function adds the block to the ast.Module at self.core_modules[0] AND adds the block to the end of the
        self.core_modules list. Any Nodes that are added to the block in future TemplateCompiler::compile_*() method
        calls are ALSO added to self.core_modules[0], because they are modifying the same object.

        The user of the TemplateCompiler instance is expected to call TemplateCompiler::close_block() when the
        end of a block is encountered.

        :param block: The block to add future expressions to
        :param start_offset: The offset for the start of the block
        :param end_offset: The offset for the end of the block
        :return:
        """
        self.add_statement(block, start_offset, end_offset)
        self.code_modules.append(block)
        self.selected_body = block.body
        self.variables.begin_scope()

    def close_block(self):
        """
        Used by a caller to signal the caller has finished parsing a block ('if', 'for', etc.).
        Removes the ast.Module from the end of the `self.code_modules` list.

        :return:
        """
        module = self.code_modules.pop()
        if not module.body:
            module.body.append(ast.Pass())

        self.variables.end_scope()
        self.selected_body = self.code_modules[-1].body

    def start_else_block(self, start_offset: int, end_offset: int):
        if not hasattr(self.code_modules[-1], "orelse"):
            raise self.syntax_error("Cannot start an else block without an if/for block", start_offset, end_offset)

        self.selected_body = self.code_modules[-1].orelse

    def finalize(self) -> ast.Module:
        assertion_error_message = (
            "There is more than one active block being parsed. "
            'Did you start compiling a block (e.g. "if", "for", etc.) '
            "and forget to call TemplateCompiler::close_block()?"
        )
        assert len(self.code_modules) == 1, assertion_error_message
        module = self.code_modules[0]
        ast.fix_missing_locations(module)
        return module

    def compile_expression(self, expr: str, offset: int) -> None:
        expression = self._compile_expression(expr, offset)
        self.append(expression, offset, offset + len(expr))

    def _compile_expression(self, expr: str, offset: int) -> ast.expr:
        """
        Compiles an expression into an Abstract Syntax Tree (AST) expression.

        The expression may contain filters, which are separated by the pipe character ('|'). Filters are applied to the
        expression from left to right. The result of the expression is passed to the first filter, and the result of
        that is passed to the second filter, and so on.

        @param expr: An expression which may contain filters
        @param offset: The offset of the start of the expression
        @return: An AST expression representing the expression
        """
        expr, *filters = expr.split("|")
        f_start_offset = offset + len(expr)

        expr, start_offset, end_offset = self.str_strip(expr, offset)
        n_expr = self._compile_node(expr, start_offset)
        self.set_location(n_expr, start_offset, end_offset)

        for filter_ in filters:
            f_start_offset += 1  # accounts for the pipe character
            n_expr = self._compile_filter(filter_, n_expr, f_start_offset)
            f_start_offset += len(filter_)

        return n_expr

    def _compile_filter(self, filter_: str, expr: ast.expr, offset: int) -> ast.expr:
        """
        Compiles a filter into an Abstract Syntax Tree (AST) expression.

        @param filter_: The filter to compile
        @param expr: The expression to apply the filter to
        @param offset: The offset of the start of the filter
        @return: An AST expression representing the filter application
        """
        filter_, start_offset, end_offset = self.str_strip(filter_, offset)
        filter_name, *args = filter_.split(":")

        if len(args) == 1:
            args = self._compile_node(f"[{args[0]}]", start_offset + len(filter_name)).elts
        elif len(args) > 1:
            raise self.syntax_error(f"Invalid Syntax", start_offset, end_offset)
        else:
            args = []

        if filter_name not in TEMPLATE_FILTERS:
            raise self.syntax_error(f"Unknown filter: {filter_name}", start_offset, start_offset + len(filter_name))

        pipe_name = f"_p_{filter_name}"
        self.variables.define(pipe_name, TEMPLATE_FILTERS[filter_name])

        args.insert(0, expr)

        n_expr: ast.Call = ast.Call(ast.Name(id=pipe_name, ctx=ast.Load()), args=args, keywords=[])
        self.set_location(n_expr, start_offset, end_offset)
        return n_expr

    def compile_legacy_expression(self, expr: str, offset: int) -> None:
        if expr[2] == "{":
            if expr[-2] == "}":
                n_expr = ast.Constant("${" + expr[3:-2] + "}", kind=None)
            else:
                n_expr = ast.Constant(expr, kind=None)
        elif expr[-2] == "}":
            n_expr = ast.Constant(expr, kind=None)
        else:
            expr, start_offset, end_offset = self.str_strip(expr[2:-1], offset + 2)
            n_expr = self._compile_node(expr, start_offset)
            self.set_location(n_expr, start_offset, end_offset)

        self.append(n_expr, offset, offset + len(expr))

    def compile_if(self, expr: str, offset: int):
        if not expr:
            raise self.syntax_error("Missing condition expression", offset, offset + len(expr))

        expr, start_offset, end_offset = self.str_strip(expr, offset + 2)
        n_expr = self._compile_node(expr, start_offset)
        block: ast.If = ast.If(n_expr, body=[], orelse=[])

        self.start_block(block, offset, end_offset)

    def compile_for(self, expr: str, offset: int):
        expr, start_offset, end_offset = self.str_strip(expr, offset)
        var_name, expr = self.split_on_whitespace(expr)

        expr, op_start_offset, _ = self.str_strip(expr, start_offset + len(var_name))
        op, container = self.split_on_whitespace(expr)

        container, c_start_offset, c_end_offset = self.str_strip(container, op_start_offset + len(op))
        var_name = var_name.strip()

        if not VAR_NAME_RE.match(var_name) or op.strip() != "in" or not container:
            raise self.syntax_error("Invalid syntax", start_offset, end_offset)

        self.variables.declare(var_name)

        block: ast.For = ast.For(
            body=[],
            orelse=[],
            target=ast.Name(id=var_name, ctx=ast.Store()),
            iter=self._compile_node(container, c_start_offset),
        )
        self.start_block(block, start_offset, end_offset)

    def compile_include(self, args: str, tag_offset: int, start_offset: int, end_offset: int) -> None:
        arg_start_offset = start_offset + 7
        indent = self.offset_to_position(tag_offset).column - 1

        node = self._compile_node(f"f({args})", arg_start_offset, -2)
        node.func = ast.Name(id=INCLUDE_VAR_NAME, ctx=ast.Load())

        has_indent = any(k.arg == "indent" for k in node.keywords)
        node.keywords.append(ast.keyword(arg="context", value=ast.Name(CONTEXT_VAR_NAME, ctx=CTX_LOAD)))
        if not has_indent:
            node.keywords.append(ast.keyword(arg="indent", value=ast.Constant(value=indent, kind=None)))

        self.set_location(node, start_offset, end_offset)
        self.set_location(node.func, start_offset, arg_start_offset)
        self.append(node, start_offset, end_offset)

    def compile_constant(self, const: str, offset: int) -> None:
        self.append(ast.Constant(value=const, kind=None), offset, offset + len(const))

    def _compile_node(self, value: str, offset: int, offset_drift: int = 0) -> ast.expr:
        lineno, offset = self.offset_to_position(offset)
        offset += offset_drift

        try:
            expr = ast.parse(value, mode="eval").body
        except SyntaxError as ex:
            if ex.lineno == 1:
                ex.offset += offset - 1

            ex.lineno += lineno - 1
            end_pos = None

            # TODO: Remove check after dropping support for Python < 3.10
            if getattr(ex, "end_lineno", None) is not None:
                if ex.end_lineno == 1:
                    ex.end_offset += offset - 1

                ex.end_lineno += lineno - 1
                end_pos = Position(ex.end_lineno, ex.end_offset)

            raise TemplateError(
                ex.msg,
                file_name=self.file_path,
                start_pos=Position(ex.lineno, ex.offset),
                end_pos=end_pos,
            ) from None

        transformer = TemplateNodeTransformer(self, lineno, offset)
        return transformer.visit(expr)

    def disallowed(self, message: str, start_position: Position, end_position: Position) -> Exception:
        return TemplateError(
            message,
            file_name=self.file_path,
            start_pos=start_position,
            end_pos=end_position,
        )

    def syntax_error(self, message: str, start_offset: int, end_offset: int) -> Exception:
        return TemplateError(
            message,
            file_name=self.file_path,
            start_pos=self.offset_to_position(start_offset),
            end_pos=self.offset_to_position(end_offset),
        )


class TemplateNodeTransformer(ast.NodeTransformer):
    def __init__(self, compiler: TemplateCompiler, initial_lineno: int, initial_offset: int):
        """
        A NodeTransformer which updates the location of nodes in an Abstract Syntax Tree. The locations are updated
        relative to the start of some source which the AST was parsed from.

        initial_lineno and initial_offset are used to update the location of nodes in the AST. The location of a node
        is updated by adding initial_lineno to the node's lineno, and adding initial_offset to the node's col_offset.
        Both initial_lineno and initial_offset are 1-based indices.

        @param compiler: The TemplateCompiler instance which is using this transformer
        @param initial_lineno: 0-based line number
        @param initial_offset: 0-based column number
        """
        self.compiler = compiler
        self.variable_refs = self.compiler.variables.refs

        self.initial_lineno = initial_lineno
        self.initial_offset = initial_offset

        self.skip_refs = False

    def update_location(self, node: ast.AST) -> None:
        if "lineno" in node._attributes:  # NOQA
            if node.lineno == 1:
                node.col_offset += self.initial_offset - 1

            if node.end_lineno == 1:
                node.end_col_offset += self.initial_offset - 1

            node.lineno += self.initial_lineno - 1
            node.end_lineno += self.initial_lineno - 1

    @staticmethod
    def node_bounds(node: ast.AST) -> tuple[Position, Position]:
        start_position = Position(node.lineno, node.col_offset + 1)
        end_position = Position(node.end_lineno, node.end_col_offset + 1)
        return start_position, end_position

    def generate_ref(self, node: ast.AST) -> bool:
        """
        Check if the node contains references to variables. If it does, add the references to the `variable_refs` field
        and return True. Else, return False.

        If self.skip_refs is True, this node is part of a branch in an AST which has already been traversed and some
        expression in this node's ancestors has been found to contain references. In this case, we do not need to check
        for references because any potential references in this node have already been added to the `variable_refs`
        field.

        @param node: The node to check for references
        @return: True if the node contains references or this self.skip_refs is True, False otherwise
        """
        if self.skip_refs:
            return True

        ref = VariableRef.from_ast_node(node)
        if ref:
            self.variable_refs[ref.name] = ref
            return True

        return False

    def generic_visit(self, node):
        self.update_location(node)
        return super().generic_visit(node)

    def visit_Attribute(self, node: ast.Attribute):
        self.update_location(node)

        if node.attr.startswith("_"):
            raise self.compiler.disallowed(f"Unknown attribute: {node.attr}", *self.node_bounds(node))

        self.skip_refs = self.generate_ref(node)
        node.value = self.visit(node.value)
        self.skip_refs = False
        return node

    def visit_Name(self, node: ast.Name):
        self.update_location(node)

        if node.id.startswith("_"):
            raise self.compiler.disallowed("Variable names cannot start with an underscore", *self.node_bounds(node))

        self.generate_ref(node)
        if node.id in self.compiler.variables:
            return node

        new_node = ast.Subscript(
            value=ast.Name(id=CONTEXT_VAR_NAME, ctx=CTX_LOAD),
            slice=ast.Constant(node.id, kind=None),
            ctx=CTX_LOAD,
        )
        ast.copy_location(new_node, node)
        return new_node

    def visit_Subscript(self, node: ast.Subscript) -> ast.AST:
        self.update_location(node)
        node.slice = self.visit(node.slice)

        self.skip_refs = self.generate_ref(node)
        node.value = self.visit(node.value)
        self.skip_refs = False
        return node



from typing import Any

from strong_opx.management.command import ConnectCommand
from strong_opx.utils.shell import shell


class Command(ConnectCommand):
    help_text = "SSH into private host inside a project/environment"
    allow_additional_args = True
    examples = [
        "strong-opx ssh --project <project> --env <env> 10.0.0.1 ...",
        "strong-opx ssh --project <project> --env <env> primary ...",
        "strong-opx ssh --project <project> --env <env> primary:2 ...",
    ]

    def shell(self, args: list[str], options: dict[str, Any]) -> None:
        shell("ssh " + " ".join(args), shell=True)



import os
from typing import TYPE_CHECKING

from strong_opx.exceptions import ImproperlyConfiguredError

if TYPE_CHECKING:
    from strong_opx.template import Context


def update_environ_hook(context: "Context"):
    for k in context:
        if k.startswith("AZURE_"):
            os.environ[k] = context[k]

    if "AZURE_SUBSCRIPTION_ID" not in context:
        raise ImproperlyConfiguredError(
            "Azure Subscription ID is not configured. Either configure that in project or environment under azure.subscription_id"
        )

    if "AZURE_RESOURCE_GROUP" not in context:
        raise ImproperlyConfiguredError(
            "Azure Resource Group is not configured. Either configure that in project or environment under azure.resource_group"
        )



import re
from typing import Collection

from colorama import Fore

from strong_opx.utils.prompt import select_prompt


class Question:
    def from_stdin(self) -> str:
        raise NotImplementedError()


class SimpleQuestion(Question):
    def __init__(self, prompt: str, validation_re: str = None):
        self.prompt = prompt

        if validation_re:
            self.validation_re = re.compile(validation_re)
        else:
            self.validation_re = None

    def from_stdin(self) -> str:
        while True:
            answer = input(f"{self.prompt}: ")
            if not self.validation_re or self.validation_re.match(answer):
                break

            print(f"{Fore.RED}Error: Must be of {self.validation_re.pattern} pattern.{Fore.RESET}")

        return answer


class ChoiceQuestion(Question):
    def __init__(self, prompt: str, choices: Collection[str]):
        self.prompt = prompt
        self.choices = choices

    def from_stdin(self) -> str:
        return select_prompt(self.prompt, self.choices)



from dataclasses import dataclass
from typing import Any, Optional

import yaml


@dataclass(frozen=True)
class Position:
    """
    Represents a position in a file. Both line and column are 1-based.
    """

    line: int
    column: Optional[int]

    def __iter__(self):
        yield self.line
        yield self.column

    def __gt__(self, other):
        if self.line > other.line:
            return True
        if self.line == other.line:
            return (self.column or 0) > (other.column or 0)
        return False

    def __lt__(self, other):
        if self.line < other.line:
            return True
        if self.line == other.line:
            return (self.column or 0) < (other.column or 0)
        return False

    def __ge__(self, other):
        return self > other or self == other

    def __le__(self, other):
        return self < other or self == other

    @classmethod
    def from_yaml_mark(cls, mark: yaml.Mark) -> "Position":
        # YAML line & column are 0-based, but we want 1-based
        return cls(line=mark.line + 1, column=mark.column + 1)

    @classmethod
    def from_offset(cls, value: str, offset: int, initial_line: int = 1, initial_col: int = 1) -> "Position":
        """
        Return the line number and column number of a character, where it resides in source value.

        :remarks: initial_line and initial_col is useful when the value is a substring of a larger value,
            and the initial line/col offset is the line/col offset of the substring in the larger value.

        :param value: source value
        :param offset: 0-based index in template
        :param initial_line: Optionally specify the initial line offset. Initial line offset is the line offset of
            the substring in the larger value.
        :param initial_col: Optionally specify the initial column offset. Initial column offset is the column offset
            of the substring in the larger value.
        :return: Position(1-based index of line, 1-based index in that line)
        """
        line = value.count("\n", 0, offset)
        col_offset = offset - value.rfind("\n", 0, offset)

        if not line:
            col_offset += initial_col - 1

        return cls(line + initial_line, col_offset)


class OpxObjectBase:
    _file_path: str = None
    _start_pos: Position = None
    _end_pos: Position = None


class OpxString(OpxObjectBase, str):
    pass


class OpxInteger(OpxObjectBase, int):
    pass


class OpxFloat(OpxObjectBase, float):
    pass


class OpxMapping(OpxObjectBase, dict):
    def __getattr__(self, item):
        try:
            return super().__getitem__(item)
        except AttributeError:
            return self[item]


class OpxSequence(OpxObjectBase, list):
    pass


def set_position(value: Any, file_path: str, start_pos: Optional[Position], end_pos: Optional[Position]) -> None:
    value._file_path = file_path
    value._start_pos = start_pos
    value._end_pos = end_pos


def set_position_from_yaml_mark(value: Any, start_mark: yaml.Mark, end_mark: yaml.Mark) -> None:
    value._file_path = start_mark.name
    value._start_pos = Position.from_yaml_mark(start_mark)
    value._end_pos = Position.from_yaml_mark(end_mark)


def get_position(value: Any) -> tuple[Optional[str], Optional[Position], Optional[Position]]:
    return getattr(value, "_file_path", None), getattr(value, "_start_pos", None), getattr(value, "_end_pos", None)



import argparse
import sys
from typing import Any

from strong_opx import yaml
from strong_opx.exceptions import CommandError
from strong_opx.management.command import ProjectCommand
from strong_opx.project import Environment
from strong_opx.vault import VaultCipher


class Command(ProjectCommand):
    help_text = "Environment variables management"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        subparsers = parser.add_subparsers(title="operation", description="Operation to execute")

        encrypt = subparsers.add_parser("encrypt", help="Encrypt value or environment variable")
        encrypt.add_argument("--value", help="Value to encrypt")
        encrypt.add_argument("--vars", nargs="*", help="Variables to encrypt")
        encrypt.set_defaults(operation="encrypt")

        decrypt = subparsers.add_parser("decrypt", help="Decrypt environment variables")
        decrypt.set_defaults(operation="decrypt")

    def handle(self, operation=None, **options: Any) -> None:
        if operation is None:
            raise CommandError("Specify operation to execute. See --help for more info")

        if operation == "encrypt":
            self.handle_encrypt(**options)
        elif operation == "decrypt":
            self.handle_decrypt(**options)

    def handle_encrypt(self, environment: Environment, **options: Any):
        if options["vars"]:
            context = environment.context

            for v in options["vars"]:
                cipher = VaultCipher.encrypt(context[v], environment.vault_secret)
                print(f"{v}: !vault |")
                for line in str(cipher).split():
                    print(" ", line)

                print()
        elif options["value"]:
            encrypted = VaultCipher.encrypt(options["value"], environment.vault_secret)
            print(str(encrypted))
        else:
            raise CommandError("Either provide --value or --vars to encrypt")

    def handle_decrypt(self, environment: Environment, **options: Any):
        decrypted_vars = environment.context
        yaml.dump(decrypted_vars.as_dict(exclude_initial=True), sys.stdout)



import datetime
from typing import Callable


def datetime_filter(value, fmt):
    if not isinstance(value, (datetime.datetime, datetime.date)):
        raise ValueError(f"Expected a datetime object, got {type(value).__name__}")

    return value.strftime(fmt)


SAFE_BUILTINS: dict[str, Callable] = {
    "range": range,
    "str": str,
}

TEMPLATE_FILTERS: dict[str, Callable] = {
    "uppercase": str.upper,
    "lowercase": str.lower,
    "titlecase": str.title,
    "strip": str.strip,
    "datetime": datetime_filter,
    "date": datetime_filter,
}



from typing import TypeVar

from strong_opx.platforms.base import Platform
from strong_opx.platforms.generic import GenericPlatform
from strong_opx.platforms.kubernetes import KubernetesPlatform

ALL_PLATFORMS = [
    GenericPlatform,
    KubernetesPlatform,
]

TPlatform = TypeVar("TPlatform", bound=Platform)



import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from strong_opx.template import Context


def update_environ_hook(context: "Context"):
    # Remove all existing gcloud environment variables to avoid conflicts
    for k in os.environ:
        if k.startswith("CLOUDSDK_"):
            del os.environ[k]

    for k in context:
        if k.startswith("CLOUDSDK_"):
            os.environ[k] = context[k]



from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union

from strong_opx.exceptions import UndefinedVariableError, VariableError
from strong_opx.template.template import Template
from strong_opx.template.variable import REF_SEP
from strong_opx.utils.tracking import OpxMapping, OpxSequence, OpxString, get_position, set_position

if TYPE_CHECKING:
    from strong_opx.template import Context

T = TypeVar("T")
T_CONTAINER = Union[dict, list]
T_CONTAINER_INDEXER = Union[str, int]
NOT_RENDERED = object()


class TreeNode:
    __slots__ = ("value", "children")

    def __init__(self, value: Any):
        self.value = value
        self.children: list[TreeNode] = []

    def __hash__(self):
        # Note this does NOT account for self.children. So, two TreeNodes with the same 'value' will have the same hash.
        # This behavior is utilized by the check_cycle() method.
        return hash(self.value)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.value})"

    def check_cycle(self, visited: dict["TreeNode", bool], in_stack: dict["TreeNode", bool]) -> Optional[list[str]]:
        # Check if node exists in the recursive stack
        if in_stack.get(self):
            return [self.value]

        # Check if node is already visited
        if visited.get(self):
            return None

        # Marking node as visited and present in recursive stack
        visited[self] = True
        in_stack[self] = True

        # Iterate for all children of node.
        for node in self.children:
            cycle_path = node.check_cycle(visited, in_stack)
            if cycle_path:
                cycle_path.append(self.value)
                return cycle_path

        # Mark 'node' to be removed from the recursive stack.
        in_stack.pop(self)
        return None


class SubstitutionBase:
    def __init__(self, ref: str):
        self.ref = ref

    def update_required_refs(self, context_refs: set) -> None:
        pass  # Does nothing by default

    def can_resolve(self, resolved_refs: set) -> bool:
        raise NotImplementedError

    def resolve(self, context: "Context") -> None:
        raise NotImplementedError

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.ref}>"


class Substitution(SubstitutionBase):
    __slots__ = ("ref", "template", "parent", "index", "required_refs")

    def __init__(self, ref: str, template: Template, parent: T_CONTAINER, index: T_CONTAINER_INDEXER):
        self.template = template
        self.required_refs = set(self.template.variables.refs)

        self.parent = parent
        self.index = index

        super().__init__(ref)

    def update_required_refs(self, context_refs: set) -> None:
        self.required_refs = {ref for ref in self.required_refs if ref.split(REF_SEP)[0] not in context_refs}

    def can_resolve(self, resolved_refs: set) -> bool:
        return self.required_refs.issubset(resolved_refs)

    def resolve(self, context: "Context") -> None:
        rendered = self.template.render(context)
        if isinstance(rendered, str):
            rendered = OpxString(rendered)
            set_position(rendered, *get_position(self.template.value))

        self.parent[self.index] = rendered


class ContainerSubstitution(SubstitutionBase):
    __slots__ = ("ref", "obj", "parent", "index")

    def __init__(self, ref: str, obj: Union[list, dict], parent: T_CONTAINER, index: T_CONTAINER_INDEXER):
        self.obj = obj
        self.parent = parent
        self.index = index

        super().__init__(ref)

    @property
    def required_refs(self):
        return set()

    def can_resolve(self, resolved_refs: set) -> bool:
        values = self.obj
        if isinstance(self.obj, dict):
            values = self.obj.values()

        return not any(v == NOT_RENDERED for v in values)

    def resolve(self, context: "Context") -> None:
        self.parent[self.index] = self.obj


class ObjectTemplate:
    __slots__ = ("context", "substitutions")

    def __init__(self, context: "Context"):
        """
        @param context: A context which is used to render values. This context is NOT modified by this class.
        """
        self.context = context
        self.substitutions: list[Union[Substitution, ContainerSubstitution]] = []

    def render(self, value: T) -> T:
        """
        Renders the given value, substituting any variables with values from the context and/or from sub-values defined
        within the value (for example, 'value' may be a dict with sub-values that reference each other).

        Note that this method does NOT honor the frozen state of the context (self.context). If 'value' contains a
        variable that is "frozen" in the context, the variable will be overwritten when rendering 'value'.
        However, it will NOT edit the variable in the context.

        For example:
        context = { 'frozen': 'frozen value' }
        value = { 'frozen': 'new frozen value', 'uses_frozen': 'blah ${frozen}' }

        would return:
        { 'frozen': 'new frozen value', 'uses_frozen': 'blah new frozen value' }

        @param value: The value to render.
        @return: The rendered value.
        """
        if isinstance(value, list):
            rendered = self.render_sequence("", value)

        elif isinstance(value, dict):
            rendered = {}
            for k, v in value.items():
                self.render_nested(ref=k, value=v, parent=rendered, index=k)

        elif isinstance(value, str):
            rendered = Template(value).render(self.context)

        else:
            return value

        context = self.context
        update_context_on_render = False

        if isinstance(rendered, dict):
            # If the value is dict, we need to update context with the rendered value so that any subsequent
            # substitutions can use the rendered value.
            update_context_on_render = True
            context = self.context.chain()
            context.update(rendered)

        self.resolve_substitutions(context, update_context_on_render)
        return rendered

    def render_nested(self, ref: str, value: Any, parent: T_CONTAINER, index: T_CONTAINER_INDEXER) -> None:
        if isinstance(value, list):
            obj = self.render_sequence(ref, value)
            parent[index] = obj
            substitution = ContainerSubstitution(ref, obj, parent, index)

        elif isinstance(value, dict):
            obj = self.render_mapping(ref, value)
            parent[index] = obj
            substitution = ContainerSubstitution(ref, obj, parent, index)

        elif isinstance(value, str):
            template = Template(value)
            substitution = Substitution(ref=ref, template=template, parent=parent, index=index)

        else:  # No need to render the value
            parent[index] = value
            return

        # Don't render the substitution yet. We need to resolve all substitutions first.
        # Because, there might be some variable used in template that is overridden
        # but context is only updated in `resolve_substitutions()` method.
        self.substitutions.append(substitution)

    def render_sequence(self, ref: str, value: list) -> list:
        parent = OpxSequence([NOT_RENDERED] * len(value))
        set_position(parent, *get_position(value))

        for i, item in enumerate(value):
            self.render_nested(f"{ref}{REF_SEP}{i}", item, parent=parent, index=i)

        return parent

    def render_mapping(self, ref: str, value: dict[str, Any]) -> dict[str, Any]:
        parent = OpxMapping()
        set_position(parent, *get_position(value))

        for k, v in value.items():
            parent[k] = NOT_RENDERED
            self.render_nested(f"{ref}{REF_SEP}{k}", v, parent=parent, index=k)

        return parent

    def resolve_substitutions(self, context: "Context", update_context_on_render: bool) -> None:
        resolved_refs = set()
        context_refs = set(context)
        substitutions = self.substitutions

        # If any context variable is referenced in the template, we need to remove it from the resolved_refs
        # so that it can be resolved again.
        for substitution in self.substitutions:
            top_ref = substitution.ref.split(REF_SEP)[0]
            if top_ref in context_refs:
                context_refs.remove(top_ref)

        # Update required_refs for each substitution based on the context_refs
        if context_refs:
            for substitution in self.substitutions:
                substitution.update_required_refs(context_refs)

        # Repeatedly iterates over the substitutions, gradually resolving values. If we end up with substitutions that
        # cannot be resolved, call handle_deadlock() which will raise the appropriate error.
        while substitutions:
            resolved = False

            i = len(substitutions) - 1
            while i >= 0:
                substitution = substitutions[i]
                if substitution.can_resolve(resolved_refs):
                    resolved_refs.add(substitution.ref)

                    substitutions.pop(i)
                    substitution.resolve(context)

                    resolved = True
                    if update_context_on_render and "." not in substitution.ref:
                        context[substitution.index] = substitution.parent[substitution.index]

                i -= 1

            if not resolved:
                # Reached deadlock while resolving substitutions. This can be either presence of a
                # circular dependency or some variable is undefined.
                self.handle_deadlock(resolved_refs)
                break

    def handle_deadlock(self, resolved_refs: set) -> None:
        """
        Compiles the unresolved variables from each substitution. If a variable does not exist, an error is raised.
        If one or more variables create a circular dependency, an error is raised.
        """
        all_nodes: dict[str, TreeNode] = {}
        defined_refs = {s.ref for s in self.substitutions} | resolved_refs

        unknowns = set()

        # loop builds trees with each tree's root being one of the string keys from load(), and each leaf of a tree
        # being a variable that root key requires for its value to be compiled/resolved.
        for substitution in self.substitutions:
            unknowns.update(substitution.required_refs - defined_refs)

            if substitution.ref in all_nodes:
                root_node = all_nodes[substitution.ref]
            else:
                root_node = TreeNode(substitution.ref)
                all_nodes[substitution.ref] = root_node

            for require in substitution.required_refs:
                if require in all_nodes:
                    node = all_nodes[require]
                else:
                    node = TreeNode(require)
                    all_nodes[require] = node

                root_node.children.append(node)

        if unknowns:
            raise UndefinedVariableError(*unknowns)

        # With the dependency trees built, walk through and find the cyclical dependencies preventing us
        # from resolving all substitutions.
        visited = {}
        in_stack = {}

        for node in all_nodes.values():
            cycle_path = node.check_cycle(visited, in_stack)
            if cycle_path:
                cycle_path.reverse()
                raise VariableError("Found circular dependency: {}".format(" -> ".join(cycle_path)), node.value)



import argparse
import logging
import os
import sys
from typing import Any

from colorama import Style

from strong_opx.exceptions import CommandError
from strong_opx.management.utils import select_environment, select_project, validate_project_name
from strong_opx.platforms import GenericPlatform
from strong_opx.project import Environment, Project
from strong_opx.providers import current_provider_error_handler

root_logger = logging.getLogger("strong_opx")
root_logger.propagate = False
root_logger.setLevel(logging.INFO)

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(
    logging.Formatter(
        f"{Style.DIM}%(asctime)s / %(levelname)s:{Style.RESET_ALL} %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
)
root_logger.addHandler(stream_handler)

logger = logging.getLogger(__name__)


def handle_command_error(e: Exception) -> None:
    if isinstance(e, CommandError):
        print(f"{e.__class__.__name__}: {e}", file=sys.stderr)
        exit(1)

    try:
        current_provider_error_handler(e)
    except CommandError as e2:
        handle_command_error(e2)


class BaseCommand:
    help_text = None
    examples = []
    allow_additional_args = False
    parse_known_args = False

    def create_parser(self, program: str = None, **kwargs: str) -> argparse.ArgumentParser:
        """
        Create and return the ``ArgumentParser`` which will be used to
        parse the arguments to this command.
        """
        epilog = ""
        help_text = self.help_text or (type(self).__doc__ or "").strip()
        if self.examples and program:
            epilog = "Examples:\n   $ {}".format("\n   $ ".join(self.examples))

        parser = argparse.ArgumentParser(
            prog=program,
            description=help_text,
            epilog=epilog,
            argument_default=argparse.SUPPRESS,
            formatter_class=type(
                "HelpFormatter", (argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter), {}
            ),
            **kwargs,
        )
        self.add_arguments(parser)
        parser.add_argument(
            "-v",
            "--verbosity",
            default=2,
            type=int,
            choices=[0, 1, 2, 3],
            help="Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output.",
        )
        parser.add_argument(
            "--traceback", action="store_true", default=False, help="Raise instead of handling known exceptions"
        )
        return parser

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        """
        Entry point for subclassed commands to add custom arguments.
        """
        pass

    def run_from_argv(self, argv: list[str]) -> None:
        parser = self.create_parser(f"{os.path.basename(argv[0])} {argv[1]}")

        additional_args = None
        if self.allow_additional_args:
            try:
                i = argv.index("--")
                additional_args = tuple(argv[i + 1 :])
                argv = argv[:i]
            except ValueError:
                additional_args = ()

        if self.parse_known_args and self.allow_additional_args:
            options, pre_additional_args = parser.parse_known_args(argv[2:])
            additional_args = tuple(pre_additional_args) + additional_args
        elif self.parse_known_args:
            options, additional_args = parser.parse_known_args(argv[2:])
            additional_args = tuple(additional_args)
        else:
            options = parser.parse_args(argv[2:])

        if options.verbosity is not None:
            root_logger.setLevel((4 - options.verbosity) * 10)

        if options.traceback:
            self.handle(**self.transform_args(options, additional_args))
        else:
            try:
                self.handle(**self.transform_args(options, additional_args))
            except Exception as e:
                handle_command_error(e)

    def transform_args(self, options: argparse.Namespace, additional_args: tuple[str, ...]) -> dict[str, Any]:
        options = vars(options)
        if additional_args is not None:
            options["additional_args"] = additional_args

        return options

    def handle(self, **options: Any) -> None:
        raise NotImplementedError()


class ProjectCommand(BaseCommand):
    require_environment = True

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--project",
            type=validate_project_name,
            help=(
                "Select project by its name. If missing, strong-opx will attempt to auto-select project based "
                "on current working directory and its parents and looking for presence of configured project"
            ),
        )
        parser.add_argument(
            "--env",
            dest="environment",
            help="Environment name. If project has only one environment that will be auto selected",
        )

    def transform_args(self, options: argparse.Namespace, additional_args: tuple[str, ...]) -> dict[str, Any]:
        options = super().transform_args(options, additional_args)
        project = options["project"] = select_project(options.get("project"))

        if self.require_environment:
            environment = options["environment"] = select_environment(project, options.get("environment"))
        else:
            environment = options.get("environment")
            if environment:
                options["environment"] = project.select_environment(environment)

        if environment:
            print(f"Using Project: {Style.BRIGHT}{project.name} [{environment.name}]{Style.RESET_ALL}")
        else:
            print(f"Using Project: {Style.BRIGHT}{project.name}{Style.RESET_ALL}")

        return options


class ConnectCommand(ProjectCommand):
    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        super().add_arguments(parser)
        parser.add_argument("hostname", help="Hostname or host group or ip of instance")

    def handle(
        self,
        environment: Environment,
        project: Project,
        hostname: str,
        **options: Any,
    ) -> None:
        platform = environment.select_platform(GenericPlatform)

        instance = platform.resolve_instance(hostname)[0]
        ssh_host = platform.get_ssh_host(instance)
        args = ["-i", project.config.ssh_key, f"{project.config.ssh_user}@{ssh_host}"]

        ssh_proxy_command = platform.ssh_proxy_command(instance.hostname)
        if ssh_proxy_command:
            args.append("-o")
            args.append(f"ProxyCommand='{ssh_proxy_command}'")

        if options.get("additional_args"):
            args = args + list(options["additional_args"])

        self.shell(args, options)

    def shell(self, args: list[str], options: dict[str, Any]) -> None:
        raise NotImplementedError()



from pydantic.dataclasses import dataclass

from strong_opx.exceptions import ImproperlyConfiguredError


@dataclass
class GCloudConfig:
    project: str = None
    compute_region: str = None

    def dict(self):
        d = {}
        if self.project:
            d["CLOUDSDK_CORE_PROJECT"] = self.project
        else:
            raise ImproperlyConfiguredError(
                "GCP Project ID is not configured. Either configure that in project or environment under gcloud.project"
            )

        if self.compute_region:
            d["CLOUDSDK_COMPUTE_REGION"] = self.compute_region
        else:
            raise ImproperlyConfiguredError(
                "GCP Compute Region is not configured. Either configure that in project or environment under gcloud.compute_region"
            )

        return d



from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel

from strong_opx.exceptions import CommandError
from strong_opx.platforms.deployments import DeploymentProvider
from strong_opx.platforms.plugins import PlatformPlugin

if TYPE_CHECKING:
    from strong_opx.project import Environment, Project
    from strong_opx.template import Context


class Platform:
    config_class: type[BaseModel]
    config_namespace: str = None

    deployment_providers: tuple[type[DeploymentProvider], ...] = ()
    plugins: dict[str, type[PlatformPlugin]] = {}

    def __init__(self, project: "Project", environment: "Environment", **kwargs: Any):
        self.project = project
        self.environment = environment

        for k, v in kwargs.items():
            setattr(self, k, v)

    def init_context(self, context: "Context") -> None:
        pass

    @classmethod
    def extract_relevant_config(cls, config: dict[str, Any]) -> Optional[dict[str, Any]]:
        values = {}

        for name, field in cls.config_class.model_fields.items():
            if name in config:
                values[name] = config[name]
            elif field.is_required():
                return None

        if values:
            return values

        return None

    @classmethod
    def from_config(cls, project: "Project", environment: "Environment", **config: Any) -> Optional["Platform"]:
        if cls.config_namespace is not None:
            platform_config = config.get(cls.config_namespace)
        else:
            platform_config = cls.extract_relevant_config(config)

        if platform_config is not None:
            platform_config = cls.config_class(**platform_config).model_dump()
            return cls(project=project, environment=environment, **platform_config)

    def plugin(self, plugin_name: str) -> "PlatformPlugin":
        plugin_class = self.plugins.get(plugin_name.lower())
        if plugin_class is None:
            raise CommandError(f"Unknown plugin: {plugin_name}")

        return plugin_class(self)



from strong_opx import __version__
from strong_opx.codegen.generator import TemplateGenerator
from strong_opx.template import Context


class ProjectGenerator(TemplateGenerator):
    template_dir = "project"

    def __init__(self, name: str, output_dir: str):
        self.name = name
        self.output_dir = output_dir
        super().__init__()

    def get_output_dir(self, context: "Context") -> str:
        return self.output_dir

    def get_context(self, context: "Context") -> "Context":
        context.update({"project_name": self.name, "strong_opx_version": __version__})
        return context



import os
from textwrap import TextWrapper
from typing import Generator, Union

from colorama import Fore, Style

from strong_opx.utils.tracking import Position, get_position


class CommandError(Exception):
    pass


class RepositoryNotFoundException(Exception):
    pass


class ComputeInstanceError(ValueError, CommandError):
    pass


class ImproperlyConfiguredError(CommandError):
    pass


class ProjectError(CommandError):
    pass


class ProjectEnvironmentError(ProjectError):
    pass


class ProcessError(CommandError):
    pass


class HelmError(CommandError):
    pass


class PluginError(CommandError):
    pass


class VaultError(CommandError):
    pass


class ErrorDetail:
    lines_for_context = 2
    __slots__ = ("error", "file_path", "start_pos", "end_pos", "hint")

    def __init__(
        self,
        error: str,
        file_path: str = None,
        start_pos: Position = None,
        end_pos: Position = None,
        hint: str = None,
    ):
        self.error = error
        self.file_path = file_path
        self.start_pos = start_pos
        self.end_pos = end_pos
        self.hint = hint

    def read_block(self, file_path: str) -> Generator[str, None, None]:
        def read_lines(start_index: int, end_index: int) -> Generator[str, None, None]:
            with open(file_path) as f:
                for _ in range(start_index - 1):
                    f.readline()

                last_line = f.readline()
                for _ in range(end_index - start_index):
                    yield last_line.rstrip()
                    last_line = f.readline()

                yield last_line.rstrip()

        lines = []

        start_line = max(1, self.start_pos.line - self.lines_for_context)
        end_line = (self.end_pos.line if self.end_pos else self.start_pos.line) + self.lines_for_context
        before_context_lines = self.start_pos.line - start_line
        n_problem_lines = (self.end_pos.line if self.end_pos else self.start_pos.line) - self.start_pos.line + 1

        lines_reader = read_lines(start_line, end_line)

        for _ in range(before_context_lines):
            line = next(lines_reader)
            if not line:
                start_line += len(lines) + 1
                lines = []
            else:
                lines.append(line)

        for _ in range(n_problem_lines):
            lines.append(next(lines_reader))

        ln_content = lines[-1]
        if self.end_pos and self.end_pos.column:
            lines[-1] = f"{ln_content[:self.end_pos.column - 1]}{Style.RESET_ALL}{ln_content[self.end_pos.column - 1:]}"
        else:
            lines[-1] += Style.RESET_ALL

        l1_content = lines[-n_problem_lines]
        if self.start_pos.column:
            lines[-n_problem_lines] = (
                f"{l1_content[:self.start_pos.column - 1]}{Style.BRIGHT}" f"{l1_content[self.start_pos.column - 1:]}"
            )

        for _ in range(self.lines_for_context):
            line = next(lines_reader)
            if not line:
                # Exhaust iterator
                list(lines_reader)
                break

            lines.append(line)

        for n, line in enumerate(lines, start_line):
            yield f"{Style.DIM}{n}:  {Style.RESET_ALL}{line}"

    def __str__(self):
        lines = self.error.splitlines()
        lines[0] = f"{Fore.RED}Error:{Fore.RESET} {Style.BRIGHT}{lines[0]}"
        lines[-1] += Style.RESET_ALL

        if self.file_path:
            from strong_opx.project import Project

            if Project.current():
                project_path = Project.current().path
            else:
                project_path = "."

            if os.path.isabs(self.file_path):
                display_path = os.path.relpath(self.file_path, project_path)
                file_path = self.file_path
            else:
                display_path = self.file_path
                file_path = os.path.join(project_path, self.file_path)

            # If file is from outside of project repo, don't show relative path
            if display_path[0] == ".":
                display_path = os.path.abspath(display_path)

            position_line = f"  in {display_path}"
            if self.start_pos:
                position_line += f" on line {self.start_pos.line}"
                if self.start_pos.column:
                    position_line += f" column {self.start_pos.column}"

            lines.append("")
            lines.append(position_line)

            if self.start_pos:
                for line in self.read_block(file_path):
                    lines.append(f"  {line}")

        if self.hint:
            lines.append("")
            for line in self.hint.splitlines():
                if line:
                    lines.extend(TextWrapper(width=80, replace_whitespace=False).wrap(line))
                else:  # Keep empty newlines
                    lines.append("")

        if len(lines) > 1:
            for i in range(len(lines)):
                lines[i] = f"{Fore.RED}│{Fore.RESET} {lines[i]}"

            lines.insert(0, f"{Fore.RED}╷{Fore.RESET}")
            lines.append(f"{Fore.RED}╵{Fore.RESET}")

        return "\n".join(lines)


class ConfigurationError(CommandError):
    __slots__ = ("errors",)

    def __init__(self, error: Union[str, ErrorDetail, list[ErrorDetail]], **kwargs):
        if isinstance(error, list):
            self.errors = error
        elif isinstance(error, ErrorDetail):
            self.errors = [error]
        else:
            self.errors = [ErrorDetail(error=error, **kwargs)]

    def __str__(self):
        lines = "\n".join(map(str, self.errors)).splitlines()
        if len(lines) > 1:
            lines.insert(0, "")

        return "\n".join(lines)


class YAMLError(ConfigurationError):
    pass


class VariableError(ConfigurationError):
    def __init__(self, message: str, var_name: str):
        self.var_name = var_name
        file_path, start_pos, end_pos = get_position(var_name)
        super().__init__(message, file_path=file_path, start_pos=start_pos, end_pos=end_pos)


class UndefinedVariableError(ConfigurationError):
    def __init__(self, *names: str):
        self.names = names

        errors = []
        for name in sorted(names):
            errors.append(ErrorDetail(f"{name} is undefined", *get_position(name)))

        super().__init__(errors)


class TemplateError(ConfigurationError):
    def __init__(self, error: str, file_name: str, **kwargs):
        super().__init__([ErrorDetail(error=error, file_path=file_name, **kwargs)])



import argparse
import os
from typing import Any

from strong_opx.exceptions import CommandError
from strong_opx.management.command import ProjectCommand
from strong_opx.platforms import GenericPlatform
from strong_opx.project import Environment, Project


class Command(ProjectCommand):
    help_text = "Run ansible playbook"
    allow_additional_args = True

    def add_arguments(self, parser: argparse.ArgumentParser):
        parser.add_argument(
            "playbook", help="Ansible playbook name (without extension) to run. Must be a file in playbooks/ directory"
        )
        parser.add_argument(
            "--host", nargs="*", dest="host_groups", metavar="GROUP_NAME", help="Specify hosts to deploy to"
        )
        parser.add_argument(
            "--keep-host-state-unchanged", action="store_true", default=False, help="Keep host state Unchanged"
        )
        super().add_arguments(parser)

    def handle(
        self,
        playbook: str,
        project: Project,
        environment: Environment,
        host_groups: list[str] = None,
        keep_host_state_unchanged: bool = False,
        **options: Any,
    ):
        platform = environment.select_platform(GenericPlatform)

        if host_groups:
            unknown_hosts = set(host_groups) - set(platform.hosts)
            if unknown_hosts:
                unknown_hosts = ", ".join(unknown_hosts)
                known_hosts = ", ".join(platform.hosts)
                raise CommandError(f"Unknown host groups: {unknown_hosts}. Choices: {known_hosts}")

        playbook_path = os.path.join(
            project.path,
            "playbooks",
            f"{playbook}.yml",
        )
        if not os.path.exists(playbook_path):
            raise CommandError(f"Unknown playbook: {playbook}")

        additional_args = ()
        if options["additional_args"]:
            additional_args += tuple(options["additional_args"])

        if keep_host_state_unchanged:
            platform.ansible_playbook(playbook_path, additional_args=additional_args, host_groups=host_groups)
        else:
            with platform.ensure_instances_are_running(host_groups):
                platform.ansible_playbook(playbook_path, additional_args=additional_args, host_groups=host_groups)



import strong_opx.template.filters  # NOQA
from strong_opx.template.context import Context
from strong_opx.template.file_template import FileTemplate
from strong_opx.template.object_template import ObjectTemplate
from strong_opx.template.template import Template
